import wgpu  # only for flags/enums

from ....resources import Texture
from ....objects import Points
from ....materials import (
    PointsMaterial,
    PointsGaussianBlobMaterial,
    PointsMarkerMaterial,
    PointsSpriteMaterial,
)

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    RenderMask,
    load_wgsl,
    nchannels_from_format,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
)


@register_wgpu_render_function(Points, PointsMaterial)
class PointsShader(BaseShader):

    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        elif color_mode == "debug":
            self["color_mode"] = "debug"
            self["color_buffer_channels"] = 0
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

        self["edge_mode"] = material.edge_mode
        self["is_sprite"] = 0  # 0, 1, 2
        if isinstance(material, PointsSpriteMaterial):
            self["is_sprite"] = 1
            if material.sprite is not None:
                self["is_sprite"] = 2  # i.e. is sprite and has a texture

        self["size_mode"] = str(material.size_mode).split(".")[-1]
        self["size_space"] = material.size_space
        self["aa"] = material.aa

        self["draw_line_on_edge"] = False
        if isinstance(material, PointsMarkerMaterial):
            self["draw_line_on_edge"] = True

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
        ]

        if self["size_mode"] == "vertex":
            bindings.append(Binding("s_sizes", rbuffer, geometry.sizes, "VERTEX"))

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] == "vertex":
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] == "vertex_map":
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )
            bindings.extend(
                self.define_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        # Process sprite texture. Note that we can *also* have a colormap for the base color.
        if self["is_sprite"] == 2:
            sprite_sampler = GfxSampler("linear", "clamp")
            if not isinstance(material.sprite, Texture):
                raise TypeError("material sprite must be a Texture")
            sprite_view = GfxTextureView(material.sprite)
            if sprite_view.view_dim != "2d":
                raise ValueError("Sprite textures must be 2D")
            fmt = to_texture_format(sprite_view.format)
            if not ("norm" in fmt or "float" in fmt):
                raise ValueError("Sprite textures must be u8norm or float")
            self["sprite_nchannels"] = len(fmt) - len(fmt.lstrip("rgba"))
            bindings += [
                Binding("s_sprite", "sampler/filtering", sprite_sampler, "FRAGMENT"),
                Binding("t_sprite", "texture/auto", sprite_view, "FRAGMENT"),
            ]

        self["shape"] = "circle"
        if isinstance(material, PointsGaussianBlobMaterial):
            self["shape"] = "gaussian"
        elif isinstance(material, PointsMarkerMaterial):
            self["shape"] = material.marker
            custom_sdf = material.custom_sdf
            if custom_sdf is None:
                # Make a nice full square to help the user better design their
                # custom SDF
                custom_sdf = "return max(abs(coord.x), abs(coord.y)) - size * 0.5;"
            self["custom_sdf"] = custom_sdf

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material

        offset, size = wobject.geometry.positions.draw_range
        offset, size = offset * 6, size * 6

        render_mask = 0
        if wobject.render_mask:
            render_mask = wobject.render_mask
        elif material.is_transparent:
            render_mask = RenderMask.transparent
        else:
            # Get what passes are needed for the color
            if self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask |= RenderMask.transparent
                else:
                    render_mask |= RenderMask.opaque
            elif self["color_mode"] == "vertex":
                if self["color_buffer_channels"] in (2, 4):
                    render_mask |= RenderMask.all
                else:
                    render_mask |= RenderMask.opaque
            elif self["color_mode"] == "vertex_map":
                if self["colormap_nchannels"] in (2, 4):
                    render_mask |= RenderMask.all
                else:
                    render_mask |= RenderMask.opaque
            elif self["color_mode"] == "debug":
                render_mask |= RenderMask.all
            else:
                raise RuntimeError(f"Unexpected color mode {self['color_mode']}")
            # Need transparency for aa
            if material.aa:
                render_mask |= RenderMask.transparent
            # More cases
            elif isinstance(material, PointsSpriteMaterial):
                if self["sprite_nchannels"] in [2, 4]:
                    render_mask |= RenderMask.transparent
                else:
                    pass  # mixed with color, so no need to OR with opaque
            elif isinstance(material, PointsMarkerMaterial):
                if material.edge_color_is_transparent:
                    render_mask |= RenderMask.transparent
                else:
                    render_mask |= RenderMask.opaque

        return {
            "indices": (size, 1, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return load_wgsl("points.wgsl")



import math
import numpy as np
import wgpu  # only for flags/enums


from ....objects import Mesh, InstancedMesh, SkinnedMesh
from ....resources import Buffer, Texture
from ....utils import normals_from_vertices
from ....materials import (
    MeshBasicMaterial,
    MeshPhongMaterial,
    MeshToonMaterial,
    MeshNormalMaterial,
    MeshNormalLinesMaterial,
    MeshSliceMaterial,
    MeshStandardMaterial,
)

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    RenderMask,
    nchannels_from_format,
    to_texture_format,
    GfxSampler,
    GfxTextureView,
    load_wgsl,
)


@register_wgpu_render_function(Mesh, MeshBasicMaterial)
class MeshShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)

        material = wobject.material
        geometry = wobject.geometry

        # Is this an instanced mesh?
        self["instanced"] = isinstance(wobject, InstancedMesh)

        # Is this a skinned mesh?
        self["use_skinning"] = isinstance(wobject, SkinnedMesh)

        # Is this a morphing mesh?
        self["use_morph_targets"] = (
            getattr(geometry, "morph_positions", None)
            or getattr(geometry, "morph_normals", None)
            or getattr(geometry, "morph_colors", None)
        )

        # Is this a wireframe mesh?
        self["wireframe"] = getattr(material, "wireframe", False)
        self["flat_shading"] = getattr(material, "flat_shading", False)

        # Lighting off in the base class
        self["lighting"] = ""
        self["receive_shadow"] = wobject.receive_shadow

        # Per-vertex color, colormap, or a plane color?
        self["colorspace"] = "srgb"

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
                self["colorspace"] = material.map.colorspace
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            self["colorspace"] = material.map.colorspace
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        elif color_mode == "face_map":
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
            self["colorspace"] = material.map.colorspace
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

    def get_bindings(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        # We're assuming the presence of an index buffer for now
        assert getattr(geometry, "indices", None)

        # Triangles or quads?
        if geometry.indices.data is not None and geometry.indices.data.shape[-1] == 4:
            self["indexer"] = 6
        else:
            self["indexer"] = 3

        # Normals. Usually it'd be given. If not, we'll calculate it from the vertices.
        if getattr(geometry, "normals", None) is not None:
            normal_buffer = geometry.normals
        else:
            if geometry.indices.data.shape[-1] == 4:
                n1 = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data[..., :3]
                )
                n2 = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data[..., [0, 2, 3]]
                )
                normal_data = (n1 + n2) / 2
            else:
                normal_data = normals_from_vertices(
                    geometry.positions.data, geometry.indices.data
                )
            normal_buffer = Buffer(normal_data)

        # Init bindings
        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_indices", rbuffer, geometry.indices, "VERTEX"),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
            Binding("s_normals", rbuffer, normal_buffer, "VERTEX"),
        ]

        # We always need texcoords, not only for colormap
        if hasattr(geometry, "texcoords") and geometry.texcoords is not None:
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )

        if hasattr(geometry, "texcoords1") and geometry.texcoords1 is not None:
            bindings.append(
                Binding("s_texcoords1", rbuffer, geometry.texcoords1, "VERTEX")
            )
            self["use_texcoords1"] = True

        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.extend(
                # todo: check uv is present and has the right shape. (introduce uv channel for texture)
                # todo: unify the logic with other maps, introduce uv channel and texture sampler
                self.define_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        if self["use_skinning"]:
            # Skinning requires skin_index and skin_weight buffers
            assert hasattr(geometry, "skin_indices") and hasattr(
                geometry, "skin_weights"
            )

            bindings.append(
                Binding("s_skin_indices", rbuffer, geometry.skin_indices, "VERTEX")
            )
            bindings.append(
                Binding("s_skin_weights", rbuffer, geometry.skin_weights, "VERTEX")
            )

            # Skinning requires a bone_matrices buffer

            skeleton = wobject.skeleton

            bindings.append(
                Binding(
                    "u_bone_matrices",
                    "buffer/uniform",
                    skeleton.bone_matrices_buffer,
                    "VERTEX",
                )
            )

        if self["use_morph_targets"]:
            morph_texture, stride, width, morph_count = getattr(
                geometry, "_gfx_morph_texture", (None, None, None, None)
            )
            if morph_texture is None:
                morph_texture, stride, width, morph_count = self._encode_morph_texture(
                    geometry
                )
                geometry._gfx_morph_texture = (
                    morph_texture,
                    stride,
                    width,
                    morph_count,
                )

            morph_target_influences = (
                wobject._morph_target_influences
            )  # the influences buffer

            if morph_texture and morph_target_influences:
                view = GfxTextureView(morph_texture, view_dim="2d-array")
                bindings.append(
                    Binding("t_morph_targets", "texture/auto", view, "VERTEX")
                )

                self["morph_targets_count"] = min(
                    morph_target_influences.nitems, morph_count
                )
                self["morph_targets_stride"] = stride
                self["morph_targets_texture_width"] = width

            else:
                self["use_morph_targets"] = False

        F = "FRAGMENT"  # noqa: N806
        sampling = "sampler/filtering"
        sampler = GfxSampler(material.map_interpolation, "repeat")
        texturing = "texture/auto"

        # todo, check uv is present and has the right shape. (introduce uv channel for texture)

        # specular map configs, only for basic and phong materials now
        if getattr(material, "specular_map", None) and not isinstance(
            material, MeshStandardMaterial
        ):
            self._check_texture(material.specular_map)
            view = GfxTextureView(material.specular_map, view_dim="2d")
            self["use_specular_map"] = True
            bindings.append(Binding("s_specular_map", sampling, sampler, F))
            bindings.append(Binding("t_specular_map", texturing, view, F))

        # set emissive_map configs, for phong and standard materials
        if getattr(material, "emissive_map", None):
            self._check_texture(material.emissive_map)
            view = GfxTextureView(material.emissive_map, view_dim="2d")
            self["use_emissive_map"] = True
            bindings.append(Binding("s_emissive_map", sampling, sampler, F))
            bindings.append(Binding("t_emissive_map", texturing, view, F))

        # set normal_map configs
        if getattr(material, "normal_map", None):
            self._check_texture(material.normal_map)
            view = GfxTextureView(material.normal_map, view_dim="2d")
            self["use_normal_map"] = True
            bindings.append(Binding("s_normal_map", sampling, sampler, F))
            bindings.append(Binding("t_normal_map", texturing, view, F))

        # set envmap configs
        if getattr(material, "env_map", None):
            self._check_texture(material.env_map, 6)
            # TODO: Support envmap not only cube, but also equirect (hdr format)
            view = GfxTextureView(material.env_map, view_dim="cube")
            bindings.append(Binding("s_env_map", sampling, sampler, F))
            bindings.append(Binding("t_env_map", texturing, view, F))

            if isinstance(material, MeshStandardMaterial):
                self["use_IBL"] = True
            elif isinstance(material, MeshBasicMaterial):
                self["use_env_map"] = True
                self["env_combine_mode"] = getattr(
                    material, "env_combine_mode", "MULTIPLY"
                )

            self["env_mapping_mode"] = getattr(
                material, "env_mapping_mode", "CUBE-REFLECTION"
            )

        # set lightmap configs
        if getattr(material, "light_map", None):
            if "use_texcoords1" not in self.kwargs:
                raise ValueError(
                    "Light map requires a second set of texture coordinates (geometry.texcoords1), but it is not present."
                )
            else:
                self._check_texture(material.light_map)
                view = GfxTextureView(material.light_map, view_dim="2d")
                self["use_light_map"] = True
                bindings.append(Binding("s_light_map", sampling, sampler, F))
                bindings.append(Binding("t_light_map", texturing, view, F))

        # set aomap configs
        if getattr(material, "ao_map", None):
            if "use_texcoords1" not in self.kwargs:
                raise ValueError(
                    "AoMap requires a second set of texture coordinates (geometry.texcoords1), but it is not present."
                )
            else:
                self._check_texture(material.ao_map)
                view = GfxTextureView(material.ao_map, view_dim="2d")
                self["use_ao_map"] = True
                bindings.append(Binding("s_ao_map", sampling, sampler, F))
                bindings.append(Binding("t_ao_map", texturing, view, F))

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(0, bindings)

        # Instanced meshes have an extra storage buffer that we add manually
        bindings1 = {}  # non-auto-generated bindings
        if self["instanced"]:
            bindings1[0] = Binding(
                "s_instance_infos", rbuffer, wobject.instance_buffer, "VERTEX"
            )

        if self["use_morph_targets"]:
            bindings1[1] = Binding(
                "u_morph_target_influences",
                "buffer/uniform",
                morph_target_influences,
                "VERTEX",
            )

        return {
            0: bindings,
            1: bindings1,
        }

    def _encode_morph_texture(self, geometry):
        morph_positions = getattr(geometry, "morph_positions", None)
        morph_normals = getattr(geometry, "morph_normals", None)
        morph_colors = getattr(geometry, "morph_colors", None)

        vetex_data_count = 0

        if morph_positions:
            vetex_data_count = 1

        if morph_normals:
            vetex_data_count = 2

        if morph_colors:
            vetex_data_count = 3

        if vetex_data_count == 0:
            return None, None, None, None

        vertice_count = geometry.positions.nitems
        total_count = vertice_count * vetex_data_count
        width = total_count
        height = 1

        max_texture_width = 4096  # TODO: use wgpu capabilities "max_texture_size"

        if width > max_texture_width:
            height = math.ceil(width / max_texture_width)
            width = max_texture_width

        morph_count = len(morph_positions or morph_normals or morph_colors or [])

        buffer = np.zeros((morph_count, height * width, 4), dtype=np.float32)

        for i in range(morph_count):
            if morph_positions:
                morph_position = morph_positions[i]
                assert (
                    len(morph_position) == vertice_count
                ), f"Morph target {i} has {len(morph_position)} vertices, expected {vertice_count}"
                morph_position = np.pad(morph_position, ((0, 0), (0, 1)), "constant")
            else:
                morph_position = np.zeros((vertice_count, 4), dtype=np.float32)

            if morph_normals:
                morph_normal = morph_normals[i]
                assert (
                    len(morph_normal) == vertice_count
                ), f"Morph normal {i} has {len(morph_normal)} vertices, expected {vertice_count}"
                morph_normal = np.pad(morph_normal, ((0, 0), (0, 1)), "constant")
            else:
                morph_normal = np.zeros((vertice_count, 4), dtype=np.float32)

            if morph_colors:
                morph_color = morph_colors[i]
                assert (
                    len(morph_color) == vertice_count
                ), f"Morph color {i} has {len(morph_colors[i])} vertices, expected {vertice_count}"
            else:
                morph_color = np.zeros((vertice_count, 4), dtype=np.float32)

            morph_data = np.stack(
                (morph_position, morph_normal, morph_color)[:vetex_data_count], axis=1
            ).reshape(-1, 4)

            buffer[i, :total_count, :] = morph_data

        return (
            Texture(buffer, dim=2, size=(width, height, morph_count)),
            vetex_data_count,
            width,
            morph_count,
        )

    def get_pipeline_info(self, wobject, shared):
        material = wobject.material

        topology = wgpu.PrimitiveTopology.triangle_list

        if material.side == "front":
            cull_mode = wgpu.CullMode.back
        elif material.side == "back":
            cull_mode = wgpu.CullMode.front
        else:  # material.side == "both"
            cull_mode = wgpu.CullMode.none

        return {
            "primitive_topology": topology,
            "cull_mode": cull_mode,
        }

    def get_render_info(self, wobject, shared):
        geometry = wobject.geometry
        material = wobject.material

        if geometry.indices.data is not None and geometry.indices.data.shape[-1] == 4:
            offset, size = geometry.indices.draw_range
            offset, size = 6 * offset, 6 * size
        else:
            offset, size = geometry.indices.draw_range
            offset, size = 3 * offset, 3 * size

        n_instances = 1
        if self["instanced"]:
            n_instances = wobject.instance_buffer.nitems

        render_mask = wobject.render_mask
        if not render_mask:
            render_mask = RenderMask.all
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask = RenderMask.transparent
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] in ("vertex", "face"):
                if self["color_buffer_channels"] in (1, 3):
                    render_mask = RenderMask.opaque
            elif self["color_mode"] in ("vertex_map", "face_map"):
                if self["colormap_nchannels"] in (1, 3):
                    render_mask = RenderMask.opaque
            elif self["color_mode"] == "normal":
                render_mask = RenderMask.opaque
            else:
                raise RuntimeError(f"Unexpected color mode {self['color_mode']}")

        return {
            "indices": (size, n_instances, offset, 0),
            "render_mask": render_mask,
        }

    def _check_texture(self, t, size2=1):
        assert isinstance(t, Texture)
        assert t.size[2] == size2
        fmt = to_texture_format(t.format)
        assert "norm" in fmt or "float" in fmt

    def get_code(self):
        return load_wgsl("mesh.wgsl")


@register_wgpu_render_function(Mesh, MeshNormalMaterial)
class MeshNormalShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["color_mode"] = "normal"
        self["colormap_dim"] = ""  # disable texture if there happens to be one


@register_wgpu_render_function(Mesh, MeshPhongMaterial)
class MeshPhongShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "phong"


@register_wgpu_render_function(Mesh, MeshToonMaterial)
class MeshToonShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "toon"

    def get_bindings(self, wobject, shared):
        result = super().get_bindings(wobject, shared)

        geometry = wobject.geometry
        material = wobject.material

        bindings = []

        F = "FRAGMENT"  # noqa: N806

        # We need uv to use the maps, so if uv not exist, ignore all maps
        if hasattr(geometry, "texcoords") and geometry.texcoords is not None:
            # Texcoords must always be nx2 since it used for all texture maps.
            nchannels = nchannels_from_format(geometry.texcoords.format)
            if not (geometry.texcoords.data.ndim == 2 and nchannels == 2):
                raise ValueError("For toon material, the texcoords must be Nx2")

            if material.gradient_map is not None:
                self._check_texture(material.gradient_map)
                view = GfxTextureView(material.gradient_map, view_dim="2d")
                self["use_gradient_map"] = True
                bindings.append(
                    Binding(
                        "s_gradient_map",
                        "sampler/filtering",
                        GfxSampler("nearest", "repeat"),
                        F,
                    )
                )  # gradient map is always nearest sampling
                bindings.append(Binding("t_gradient_map", "texture/auto", view, F))

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(2, bindings)

        # Update result
        result[2] = bindings
        return result


@register_wgpu_render_function(Mesh, MeshStandardMaterial)
class MeshStandardShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["lighting"] = "pbr"

    def get_bindings(self, wobject, shared):
        result = super().get_bindings(wobject, shared)

        geometry = wobject.geometry
        material = wobject.material

        bindings = []

        F = "FRAGMENT"  # noqa: N806
        sampling = "sampler/filtering"
        sampler = GfxSampler(material.map_interpolation, "repeat")
        texturing = "texture/auto"

        # We need uv to use the maps, so if uv not exist, ignore all maps
        if hasattr(geometry, "texcoords") and geometry.texcoords is not None:
            # Texcoords must always be nx2 since it used for all texture maps.
            nchannels = nchannels_from_format(geometry.texcoords.format)
            if not (geometry.texcoords.data.ndim == 2 and nchannels == 2):
                raise ValueError("For standard material, the texcoords must be Nx2")

            if material.roughness_map is not None:
                self._check_texture(material.roughness_map)
                view = GfxTextureView(material.roughness_map, view_dim="2d")
                self["use_roughness_map"] = True
                bindings.append(Binding("s_roughness_map", sampling, sampler, F))
                bindings.append(Binding("t_roughness_map", texturing, view, F))

            if material.metalness_map is not None:
                self._check_texture(material.metalness_map)
                view = GfxTextureView(material.metalness_map, view_dim="2d")
                self["use_metalness_map"] = True
                bindings.append(Binding("s_metalness_map", sampling, sampler, F))
                bindings.append(Binding("t_metalness_map", texturing, view, F))

        # Define shader code for binding
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(2, bindings)

        # Update result
        result[2] = bindings
        return result


@register_wgpu_render_function(Mesh, MeshNormalLinesMaterial)
class MeshNormalLinesShader(MeshShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["color_mode"] = "uniform"
        self["lighting"] = ""
        self["wireframe"] = False

    def get_pipeline_info(self, wobject, shared):
        d = super().get_pipeline_info(wobject, shared)
        d["primitive_topology"] = wgpu.PrimitiveTopology.line_list
        return d

    def get_render_info(self, wobject, shared):
        # We directly look at the vertex data, so geometry.indices.draw_range is ignored.
        d = super().get_render_info(wobject, shared)
        d["indices"] = wobject.geometry.positions.nitems * 2, d["indices"][1]
        return d

    def get_code(self):
        return load_wgsl("mesh_normal_lines.wgsl")


@register_wgpu_render_function(Mesh, MeshSliceMaterial)
class MeshSliceShader(BaseShader):
    """Shader for rendering mesh slices."""

    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)

        material = wobject.material
        geometry = wobject.geometry

        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        elif color_mode == "face_map":
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

    def get_bindings(self, wobject, shared):
        # It would technically be possible to implement colormapping or
        # per-vertex colors, but its a tricky dance to get the per-vertex
        # data (e.g. texcoords) into a varying. And because the visual
        # result is a line, its likely that in most use-cases a uniform
        # color is preferred anyway. So for now we don't implement that.

        geometry = wobject.geometry
        material = wobject.material

        # Init uniform bindings
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
        ]

        # We're assuming the presence of an index buffer for now
        assert getattr(geometry, "indices", None)

        # Init storage buffer bindings
        rbuffer = "buffer/read_only_storage"
        bindings.append(Binding("s_indices", rbuffer, geometry.indices, "VERTEX"))
        bindings.append(Binding("s_positions", rbuffer, geometry.positions, "VERTEX"))

        # Bindings for color
        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.extend(
                self.define_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        # Let the shader generate code for our bindings
        bindings = {i: binding for i, binding in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material  # noqa
        geometry = wobject.geometry

        offset, size = geometry.indices.draw_range
        offset, size = offset * 6, size * 6

        render_mask = 0
        if wobject.render_mask:
            render_mask = wobject.render_mask
        elif material.is_transparent:
            render_mask = RenderMask.transparent
        else:
            # Get what passes are needed for the color
            if self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask |= RenderMask.transparent
                else:
                    render_mask |= RenderMask.opaque
            elif self["color_mode"] in ("vertex", "face"):
                if self["color_buffer_channels"] in (1, 3):
                    render_mask |= RenderMask.opaque
                else:
                    render_mask |= RenderMask.all
            elif self["color_mode"] in ("vertex_map", "face_map"):
                if self["colormap_nchannels"] in (1, 3):
                    render_mask |= RenderMask.opaque
                else:
                    render_mask |= RenderMask.all
            else:
                raise RuntimeError(f"Unexpected color mode {self['color_mode']}")

        return {
            "indices": (size, 1, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return load_wgsl("mesh_slice.wgsl")



"""Test that data can be accessed.
"""

import os
from pygfx.utils import get_resources_dir
from pygfx.renderers.wgpu import load_wgsl


def test_can_load_wgsl():
    wgsl = load_wgsl("line.wgsl")
    assert "fn vs_main(" in wgsl


def test_can_load_fonts():
    filename = os.path.join(get_resources_dir(), "NotoSans-Regular.ttf")
    assert os.path.isfile(filename)



from ._perspective import PerspectiveCamera


class OrthographicCamera(PerspectiveCamera):
    """An orthographic camera, useful for orthograpic views and 2D content.

    Technically, this is a PerspectiveCamera with the fov locked to zero.

    Parameters
    ----------
    width: float
        The (minimum) width of the view-cube. The actual view
        may be wider if the viewport is relatively wide.
    height: float
        The (minimum) height of the view-cube. The actual view
        may be taller if the viewport is relatively high.
    zoom: float
        An additional zoom factor, equivalent to attaching a zoom lens.
    maintain_aspect: bool
        Whether the aspect ration is maintained as the window size changes.
        Default True. If false, the dimensions are stretched to fit the window.
    depth_range: 2-tuple
        The values for the near and far clipping planes. If not given
        or None, the clip planes will be calculated automatically based
        on the fov, width, and height.
    """

    _fov_range = 0, 0

    def __init__(
        self, width=1, height=1, *, zoom=1, maintain_aspect=True, depth_range=None
    ):
        super().__init__(
            0,
            None,
            width=width,
            height=height,
            zoom=zoom,
            maintain_aspect=maintain_aspect,
            depth_range=depth_range,
        )



import numpy as np
import pylinalg as la
from time import perf_counter_ns
import weakref
import functools

from typing import Tuple, Union


PRECISION_EPSILON = 1e-7


if int(np.__version__.split(".")[0]) >= 2:
    mat_inv = np.linalg.inv
else:
    # Avoid cpu's spinning at 300%, see issue #763
    mat_inv = np.linalg.pinv


class cached:  # noqa: N801
    """Cache for computed properties.

    This descriptor implements a minimal timestamp-based cache for computed
    properties. The value of the property is computed using ``compute_fn`` and
    the result is cached until ``obj.last_modified`` advances. At this point the
    value of the computed property is recomputed upon the next read.

    """

    def __init__(self, compute_fn=None) -> None:
        self.compute_fn = compute_fn
        self.name = None

    def __set_name__(self, clazz, name) -> None:
        self.name = f"_{name}_cache"

    def __get__(self, instance: "AffineBase", clazz=None) -> np.ndarray:
        if instance is None:
            return self

        if not hasattr(instance, self.name):
            cache = (instance.last_modified, self.compute_fn(instance))
            setattr(instance, self.name, cache)
        else:
            cache = getattr(instance, self.name)

        if instance.last_modified > cache[0]:
            cache = (instance.last_modified, self.compute_fn(instance))
            setattr(instance, self.name, cache)

        return cache[1]


class callback:  # noqa: N801
    """Weakref support for AffineTransform callbacks.

    This decorator replaces the instance reference of a class method (self) with
    a weakref to the instance. It also dynamically resolves this weakref
    so that users don't have to.

    The use-case for it is to avoid the circular reference that happens when a
    WorldObject registers a callback to keep it's local transform in sync with
    its world transform. This speeds up garbage collection as WorldObjects will
    get removed once the last reference is deleted instead of waiting for GC's
    cycle detection phase.

    """

    def __init__(self, callback_fn) -> None:
        self.callback_fn = callback_fn

    def __get__(self, instance, clazz=None):
        if instance is None:
            return self

        weak_instance = weakref.ref(instance)

        @functools.wraps(self.callback_fn)
        def inner(*args, **kwargs):
            if weak_instance() is None:
                return

            return self.callback_fn(weak_instance(), *args, **kwargs)

        return inner


class AffineBase:
    """Base class for affine transformations.

    .. warning::
        In-place updates of slices of properties, e.g. ``transform.position[1] =
        42`` have no effect due to limitations of the python programming
        language and our decision to have the properties return pure numpy arrays.

    This class implements basic getters and setters for the various properties
    of an affine transformation used in pygfx. If you are looking for
    `obj.local.<something>` or `obj.world.<something>` it is probably defined
    here.

    The class further implements a basic callback system that will eagerly
    inform callees whenever the transform's state changes. Callees register
    callbacks using the ``callback_id = transform.on_update(...)`` method and -
    if the callee is a class - may optionally choose to decorate the callback
    method with the ``@callback`` descriptor defined above. This will turn the
    callback into a weakref and allow the callee to be garbage collected more
    swiftly. After registration callees can remove the callback by calling
    ``transform.remove_callback`` and passing the callback. Callees can also
    pass the ``callback_id`` returned when the callback was first registered
    (useful e.g. if the callback was a lambda).

    It also implements a basic caching mechanism that keeps computed properties
    around until the underlying transform changes. This makes use of the
    `@cached` descriptor defined above. The descriptor expects a `last_modified`
    property on the consuming class which is used as a monotonously increasing
    timestamp/counter to indicate if and when a cached value has become invalid.
    Once invalid, cached values are updated upon the next read/get meaning that
    they are updated lazily.

    Parameters
    ----------
    reference_up : ndarray, [3]
        The direction of the reference_up vector expressed in the target frame.
        It indicates neutral tilt and is used by the axis properties (right, up,
        forward) to maintain a common level of rotation around an axis when it
        is updated by it's setter. By default, it points along the Y-axis.
    is_camera_space : bool
        If True, the transform represents a camera space which means that it's
        ``forward`` and ``right`` directions are inverted.

    Notes
    -----
    Subclasses need to define and implement ``last_modified`` for the caching
    machnism to work correctly. Check out existing subclasses for an example of
    how this might look like.

    All properties are **expressed in the target frame**, i.e., they use the
    target's basis, unless otherwise specified.

    """

    last_modified: int

    def __init__(self, *, reference_up=(0, 1, 0), is_camera_space=False):
        self.update_callbacks = {}
        self.is_camera_space = int(is_camera_space)
        self._reference_up = np.asarray(reference_up, dtype=float)
        self._scaling_signs = np.asarray([1, 1, 1], dtype=float)

    @property
    def matrix(self):
        """Affine matrix describing this transform.

        ``vec_target = matrix @ vec_source``.

        """
        raise NotImplementedError()

    @cached
    def _inverse_matrix(self) -> np.ndarray:
        return mat_inv(self.matrix)

    @property
    def scaling_signs(self):
        return self._scaling_signs

    @cached
    def _decomposed(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        try:
            decomposed = la.mat_decompose(self.matrix, scaling_signs=self.scaling_signs)
        except ValueError:
            # the matrix has been set manually
            # and so there is no scaling component to preserve
            # any decomposed scaling is acceptable
            decomposed = la.mat_decompose(self.matrix)

        # Stop the user from accidentally writing the temporary arrays
        # that we return from these operations
        # https://github.com/pygfx/pygfx/issues/651
        for m in decomposed:
            m.flags.writeable = False

        return decomposed

    @cached
    def _directions(self):
        # Note: forward_is_minus_z indicates the camera frame
        directions = (2 * self.is_camera_space - 1, 1, 1 - 2 * self.is_camera_space)

        axes_source = np.diag(directions)
        axes_target = la.vec_transform(axes_source, self.matrix)
        origin_target = la.vec_transform((0, 0, 0), self.matrix)

        directions = axes_target - origin_target
        # Stop the user from accidentally writing the temporary arrays
        # that we return from these operations
        # https://github.com/pygfx/pygfx/issues/651
        directions.flags.writeable = False
        return directions

    @cached
    def _direction_components(self):
        return (*self._directions,)

    @cached
    def _rotation_matrix(self):
        rotation = la.mat_from_quat(self._decomposed[1])

        # Stop the user from accidentally writing the temporary arrays
        # that we return from these operations
        # https://github.com/pygfx/pygfx/issues/651
        rotation.flags.writeable = False
        return rotation

    @cached
    def _euler(self):
        euler = la.quat_to_euler(self._decomposed[1])

        # Stop the user from accidentally writing the temporary arrays
        # that we return from these operations
        # https://github.com/pygfx/pygfx/issues/651
        euler.flags.writeable = False
        return euler

    def flag_update(self):
        """Signal that this transform has updated."""
        for callback in list(self.update_callbacks.values()):
            callback(self)

    def on_update(self, callback) -> int:
        """Subscribe to updates of this transform.

        The provided callback will be executed after this transform has updated
        using the signature ``callback(self)``, i.e., it is passed a reference to
        this transform.

        Parameters
        ----------
        callback : Callable
            The callback to be executed after an update.

        Returns
        -------
        callback_id : int
            A ID to uniquely identify this callback and allow unsubscribing.

        """
        callback_id = id(callback)
        self.update_callbacks[callback_id] = callback

        return callback_id

    def remove_callback(self, ref) -> None:
        """Unsubscribe from updates of this transform.

        Parameters
        ----------
        ref : int, Callable
            The callback (or callback_id) to unsubscribe.

        """

        if isinstance(ref, int):
            callback_id = ref
        else:
            callback_id = id(ref)

        self.update_callbacks.pop(callback_id, None)

    @property
    def inverse_matrix(self) -> np.ndarray:
        """Inverse of this transform as affine matrix.

        ``vec_source = inverse_matrix @ vec_target``

        """
        return self._inverse_matrix

    @property
    def position(self) -> np.ndarray:
        """The origin of source."""
        return self._decomposed[0]

    @property
    def rotation(self) -> np.ndarray:
        """The orientation of source as a quaternion."""
        return self._decomposed[1]

    @property
    def rotation_matrix(self) -> np.ndarray:
        """The orientation of source as a rotation matrix."""
        return self._rotation_matrix

    @property
    def euler(self) -> np.ndarray:
        """The orientation of source as XYZ euler angles."""
        return self._euler

    @property
    def euler_x(self) -> float:
        """The X component of source's orientation as XYZ euler angles."""
        return self._euler[0]

    @property
    def euler_y(self) -> float:
        """The Y component of source's orientation as XYZ euler angles."""
        return self._euler[1]

    @property
    def euler_z(self) -> float:
        """The Z component of source's orientation as XYZ euler angles."""
        return self._euler[2]

    @property
    def scale(self) -> np.ndarray:
        """The scale of source."""
        return self._decomposed[2]

    @property
    def reference_up(self) -> np.ndarray:
        """The zero-tilt reference vector used for the direction setters."""
        return self._reference_up

    @property
    def x(self) -> float:
        """The X component of source's position."""
        return self.position[0]

    @property
    def y(self) -> float:
        """The Y component of source's position."""
        return self.position[1]

    @property
    def z(self) -> float:
        """The Z component of source's position."""
        return self.position[2]

    @property
    def scale_x(self) -> float:
        """The X component of source's scale."""
        return self.scale[0]

    @property
    def scale_y(self) -> float:
        """The Y component of source's scale."""
        return self.scale[1]

    @property
    def scale_z(self) -> float:
        """The Z component of source's scale."""
        return self.scale[2]

    @property
    def right(self):
        """The right direction of source."""
        return self._direction_components[0]

    @property
    def up(self):
        """The up direction of source."""
        return self._direction_components[1]

    @property
    def forward(self):
        """The forward direction of source."""
        return self._direction_components[2]

    @position.setter
    def position(self, value):
        self.matrix = la.mat_compose(value, self.rotation, self.scale)

    @rotation.setter
    def rotation(self, value):
        self.matrix = la.mat_compose(self.position, value, self.scale)

    @euler.setter
    def euler(self, value):
        self.rotation = la.quat_from_euler(value)

    @rotation_matrix.setter
    def rotation_matrix(self, value):
        self.rotation = la.quat_from_mat(value)

    @euler_x.setter
    def euler_x(self, value):
        self.euler = (value, 0, 0)

    @euler_y.setter
    def euler_y(self, value):
        self.euler = (0, value, 0)

    @euler_z.setter
    def euler_z(self, value):
        self.euler = (0, 0, value)

    @scale.setter
    def scale(self, value):
        m = la.mat_compose(self.position, self.rotation, value)
        self._scaling_signs = np.sign(value)
        self.matrix = m

    @reference_up.setter
    def reference_up(self, value):
        self._reference_up = la.vec_normalize(value)
        self.flag_update()

    @x.setter
    def x(self, value):
        _, y, z = self.position
        self.position = (value, y, z)

    @y.setter
    def y(self, value):
        x, _, z = self.position
        self.position = (x, value, z)

    @z.setter
    def z(self, value):
        x, y, _ = self.position
        self.position = (x, y, value)

    @scale_x.setter
    def scale_x(self, value):
        _, y, z = self.scale
        self.scale = (value, y, z)

    @scale_y.setter
    def scale_y(self, value):
        x, _, z = self.scale
        self.scale = (x, value, z)

    @scale_z.setter
    def scale_z(self, value):
        x, y, _ = self.scale
        self.scale = (x, y, value)

    @right.setter
    def right(self, value):
        value = la.vec_normalize(value)

        if np.allclose(value, (0, 0, 0)):
            raise ValueError("A coordinate axis can't point at its origin.")

        if np.linalg.norm(np.cross(value, self.reference_up)) == 0:
            # target and reference_up are parallel
            rotation = la.quat_from_vecs(self.forward, value)
        else:
            matrix = la.mat_look_at((0, 0, 0), value, self.reference_up)
            rotation = la.quat_from_mat(matrix)

        if self.is_camera_space:
            part2 = la.quat_from_axis_angle((0, 1, 0), -np.pi / 2)
        else:
            part2 = la.quat_from_axis_angle((0, 1, 0), np.pi / 2)

        self.rotation = la.quat_mul(rotation, part2)

    @up.setter
    def up(self, value):
        value = la.vec_normalize(value)

        if np.allclose(value, (0, 0, 0)):
            raise ValueError("A coordinate axis can't point at its origin.")

        if np.linalg.norm(np.cross(value, self.reference_up)) == 0:
            # target and reference_up are parallel
            rotation = la.quat_from_vecs(self.forward, value)
        else:
            matrix = la.mat_look_at((0, 0, 0), value, self.reference_up)
            rotation = la.quat_from_mat(matrix)

        part2 = la.quat_from_axis_angle((1, 0, 0), np.pi / 2)
        self.rotation = la.quat_mul(rotation, part2)

    @forward.setter
    def forward(self, value):
        value = la.vec_normalize(value)

        if np.allclose(value, (0, 0, 0)):
            raise ValueError("A coordinate axis can't point at its origin.")

        if np.linalg.norm(np.cross(value, self.reference_up)) < PRECISION_EPSILON:
            # target and reference_up are parallel
            if np.linalg.norm(np.cross(self.forward, value)) < PRECISION_EPSILON:
                # target and forward are parallel, no rotation needed
                return
            rotation = la.quat_from_vecs(self.forward, value)
        elif self.is_camera_space:
            matrix = la.mat_look_at((0, 0, 0), -value, self.reference_up)
            rotation = la.quat_from_mat(matrix)
        else:
            matrix = la.mat_look_at((0, 0, 0), value, self.reference_up)
            rotation = la.quat_from_mat(matrix)

        self.rotation = rotation

    def __array__(self, dtype=None):
        return self.matrix.astype(dtype)


class AffineTransform(AffineBase):
    """A single affine transform.

    Parameters
    ----------
    matrix : ndarray, [4, 4]
        The affine matrix used to back this transform. If None, a new diagonal
        matrix will be initialized.
    position : ndarray, [3]
        The position of this transform expressed in the target frame. This will
        overwrite the position component of ``matrix`` if present.
    rotation : ndarray, [4]
        The rotation quaternion of this transform expressed in the target frame.
        This will overwrite the rotation component of ``matrix`` if present.
    scale : ndarray, [3]
        The per-axis scale of this transform expressed in the target frame. This
        will overwrite the scale component of ``matrix`` if present.
    reference_up : ndarray, [3]
        The direction of the reference_up vector expressed in the target frame.
        It indicates neutral tilt and is used by the axis properties (right, up,
        forward) to maintain a common level of rotation around an axis when it
        is updated by it's setter. By default, it points along the Y-axis.
    is_camera_space : bool
        If True, the transform represents a camera space which means that it's
        ``forward`` and ``right`` directions are inverted.

    Notes
    -----
    The transform class "wraps" the provided ``matrix`` and shares its buffer.
    This is useful when optimizing performance, as it is possible to wrap a view
    into an (aligned) buffer of multiple transformation matrices or,
    alternatively, to directly wrap a uniform buffer. Updates to the transform
    will then directly modify this matrix which speeds up computation by
    avoiding copies or exploiting data alignment.

    When updating the underlying matrix in-place these updates will not
    propagate via the transform's callback system nor will they invalidate
    existing caches. To inform the transform of these updates call
    ``transform.flag_update()``, which will trigger both callbacks and cache
    invalidation.

    See Also
    --------
    AffineBase
        Base class defining various useful properties for this transform.

    """

    def __init__(
        self,
        matrix=None,
        /,
        *,
        position=None,
        rotation=None,
        scale=None,
        reference_up=(0, 1, 0),
        is_camera_space=False,
    ) -> None:
        super().__init__(reference_up=reference_up, is_camera_space=is_camera_space)
        self.last_modified = perf_counter_ns()

        if matrix is None:
            matrix = np.eye(4, dtype=float)

        self.untracked_matrix = np.asarray(matrix)

        if position is not None:
            self.position = position

        if rotation is not None:
            self.rotation = rotation

        if scale is not None:
            self.scale = scale

    def flag_update(self):
        self.last_modified = perf_counter_ns()
        super().flag_update()

    @property
    def matrix(self):
        view_array = self.untracked_matrix.view()
        # Stop the user from accidentally writing the temporary arrays
        # that we return from these operations
        # https://github.com/pygfx/pygfx/issues/651
        view_array.flags.writeable = False
        return view_array

    @matrix.setter
    def matrix(self, value):
        self.untracked_matrix[:] = value
        self.flag_update()

    def __matmul__(self, other):
        if isinstance(other, AffineTransform):
            return AffineTransform(
                self.matrix @ other.matrix, is_camera_space=self.is_camera_space
            )

        return np.asarray(self) @ other

    def __array__(self, dtype=None):
        return self.untracked_matrix.astype(dtype, copy=False)


class RecursiveTransform(AffineBase):
    """A transform that may be preceded by another transform.

    This transform behaves semantically identical to an ordinary
    ``AffineTransform`` (same properties), except that users may define a
    ``parent`` transform which precedes the ``matrix`` used by the ordinary
    ``AffineTransform``. The resulting ``RecursiveTransform`` then controls the
    total transform that results from combinign the two transforms via::

        recursive_transform = parent @ matrix

    In other words, the source frame of ``RecursiveTransform`` is the source
    frame of ``matrix`` and the target frame of ``RecursiveTransform`` is the
    target frame of ``parent``. Implying that ``RecursiveTransform``'s
    properties are given in the target frame.

    The use case for this class is to allow getting and setting of properties of
    a WorldObjects world transform, i.e., it implements ``WorldObject.world``.

    Under the hood, this transform wraps another transform (passed in as
    ``matrix``), similar to how ``AffineTransform`` wraps a numpy array. It can
    also be initialized from a numpy array, in which case provided matrix is
    first wrapped into a ``AffineTransform`` which is then wrapped by this
    class. Setting properties of ``RecursiveTransform`` will then internally
    transform the new values into ``matrix`` target frame and then set the
    obtained value on the wrapped transform. This means that the ``parent``
    transform is not affected by changes made to this transform.

    Further, this transform monitors ``parent`` for changes (via a callback) and
    will update (invalidate own caches and trigger callbacks) whenever the
    parent updates. This allows propagating updates from the parent to its
    children, e.g., to update a child's world transform when it's parent changes
    position.

    Parameters
    ----------
    matrix : AffineBase, ndarray [4, 4]
        The base transform that will be wrapped by this transform.
    parent : AffineBase, optional
        The parent transform that precedes the base transform.
    reference_up : ndarray, [3]
        The direction of the reference_up vector expressed in the target
        frame. It is used by the axis properties (right, up, forward)
        to maintain a common level of rotation around an axis when it
        is updated by it's setter. By default, it points along the
        positive Y-axis.
    is_camera_space : bool
        If True, the transform represents a camera space which means that it's
        ``forward`` and ``right`` directions are inverted.

    Notes
    -----
    Since ``parent`` is optional, this transform can also be used to create a
    synced copy of an existing transform similar to how a view works in numpy.

    See Also
    --------
    AffineBase
        Base class defining various useful properties for this transform.

    """

    def __init__(
        self,
        matrix: Union[np.ndarray, AffineBase],
        /,
        *,
        parent=None,
        reference_up=None,
        is_camera_space=False,
    ) -> None:
        super().__init__(is_camera_space=is_camera_space)
        self._parent = None
        self.own = None
        self.last_modified = perf_counter_ns()

        if isinstance(matrix, AffineBase):
            self.own = matrix
        else:
            self.own = AffineTransform(matrix, is_camera_space=is_camera_space)

        if parent is None:
            self._parent = AffineTransform()
        else:
            self._parent = parent

        if reference_up is None:
            # use own's reference_up
            reference_up = la.vec_transform(self.own.reference_up, self.parent.matrix)
            self._reference_up = la.vec_normalize(reference_up)
        else:
            # use given reference_up (and sync own)
            self._reference_up = la.vec_normalize(reference_up)
            self._propagate_reference_up()

        self.parent.on_update(self._parent_updated)
        self.own.on_update(self._own_updated)

    def flag_update(self):
        self.last_modified = perf_counter_ns()
        super().flag_update()

    @callback
    def _parent_updated(self, parent: AffineBase):
        self._propagate_reference_up()
        self.flag_update()

    def _propagate_reference_up(self):
        new_ref = la.vec_transform(self._reference_up, self._parent.inverse_matrix)
        origin = la.vec_transform((0, 0, 0), self._parent.inverse_matrix)
        self.own._reference_up = la.vec_normalize(new_ref - origin)

    @callback
    def _own_updated(self, own: AffineBase):
        new_ref = la.vec_transform(own.reference_up, self.parent.matrix)
        origin = self.parent.position
        self._reference_up = la.vec_normalize(new_ref - origin)
        self.flag_update()

    @AffineBase.reference_up.setter
    def reference_up(self, value):
        self._reference_up = la.vec_normalize(value)
        self._propagate_reference_up()
        self.flag_update()

    @property
    def parent(self) -> AffineBase:
        """The transform that precceeds the own/local transform."""
        return self._parent

    @parent.setter
    def parent(self, value):
        self.parent.remove_callback(self._parent_updated)

        if value is None:
            self._parent = AffineTransform()
        else:
            self._parent = value

        self._propagate_reference_up()
        self.parent.on_update(self._parent_updated)
        self.flag_update()

    @cached
    def _matrix(self):
        return self._parent.matrix @ self.own.matrix

    @property
    def matrix(self):
        return self._matrix

    @matrix.setter
    def matrix(self, value):
        self.own.matrix = self._parent.inverse_matrix @ value

    def __matmul__(self, other):
        if isinstance(other, AffineBase):
            return RecursiveTransform(other, parent=self)
        else:
            return np.asarray(self) @ other

    @cached
    def scaling_signs(self):
        return self._parent.scaling_signs * self.own.scaling_signs



""" pygfx line shader. See line.wgsl for details.
"""

import wgpu  # only for flags/enums
import numpy as np
import pylinalg as la

from ....utils import array_from_shadertype
from ....resources import Buffer
from ....objects import Line
from ....materials._line import (
    LineMaterial,
    LineSegmentMaterial,
    LineArrowMaterial,
    LineThinMaterial,
    LineThinSegmentMaterial,
    LineDebugMaterial,
)

from .. import (
    register_wgpu_render_function,
    BaseShader,
    Binding,
    RenderMask,
    load_wgsl,
    nchannels_from_format,
)


renderer_uniform_type = dict(last_i="i4")


@register_wgpu_render_function(Line, LineMaterial)
class LineShader(BaseShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        material = wobject.material
        geometry = wobject.geometry

        self["line_type"] = "line"
        self["dashing"] = False
        self["thickness_space"] = material.thickness_space
        self["aa"] = material.aa
        self["debug"] = False

        # Handle color
        color_mode = str(material.color_mode).split(".")[-1]
        if color_mode == "auto":
            if material.map is not None:
                self["color_mode"] = "vertex_map"
                self["color_buffer_channels"] = 0
            else:
                self["color_mode"] = "uniform"
                self["color_buffer_channels"] = 0
        elif color_mode == "uniform":
            self["color_mode"] = "uniform"
            self["color_buffer_channels"] = 0
        elif color_mode == "vertex":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "vertex"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "face":
            nchannels = nchannels_from_format(geometry.colors.format)
            self["color_mode"] = "face"
            self["color_buffer_channels"] = nchannels
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
        elif color_mode == "vertex_map":
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
            self["color_mode"] = "vertex_map"
            self["color_buffer_channels"] = 0
        elif color_mode == "face_map":
            if material.map is None:
                raise ValueError(f"Cannot apply colormap is no material.map is set.")
            self["color_mode"] = "face_map"
            self["color_buffer_channels"] = 0
        else:
            raise RuntimeError(f"Unknown color_mode: '{color_mode}'")

        # Optimization: when the line is opaque, has a uniform color, and no dashing,
        # it can be rendered pretty safely without joins. I *think* this is faster,
        # because a lot of logic related joins becomes simpler. However, the miters
        # result in extra fragments that need to be processed, so we'd need to do
        # some benchmarks to be sure.
        # if (
        #     self["color_mode"] == "uniform"
        #     and not self["dashing"]
        #     and not material.is_transparent
        #     and not material.color_is_transparent
        # ):
        #     # self["line_type"] = "quickline"

        # Handle dashing
        if material.dash_pattern:
            # Set dash props
            self["dashing"] = True
            self["dash_pattern"] = tuple(wobject.material.dash_pattern)
            self["dash_count"] = len(wobject.material.dash_pattern) // 2
            # For line segments we can calculate the distance between nodes in the shader.
            # For normal lines, we need a cumulative distance.
            if not isinstance(material, LineSegmentMaterial):
                self.needs_bake_function = True
                self._positions_hash = None
                self.line_distance_buffer = Buffer(
                    np.zeros((geometry.positions.nitems,), np.float32)
                )

    def bake_function(self, wobject, camera, logical_size):
        # Prepare
        positions_buffer = wobject.geometry.positions
        r_offset, r_size = positions_buffer.draw_range

        # Prepare arrays
        positions_array = positions_buffer.data[r_offset : r_offset + r_size]
        distance_array = self.line_distance_buffer.data[r_offset : r_offset + r_size]

        # Get vertices in the appropriate coordinate frame
        if wobject.material.thickness_space == "model":
            # Skip this step if the position data has not changed
            positions_hash = (id(positions_buffer), positions_buffer.rev)
            if positions_hash == self._positions_hash:
                return
            self._positions_hash = positions_hash
            vertex_array = positions_array
        else:
            # Prep
            finites = np.isfinite(positions_array).all(axis=1)
            has_non_finites = not finites.all()
            if has_non_finites:
                positions_array_sub = positions_array[finites, :]
            else:
                positions_array_sub = positions_array
            # Transform
            if wobject.material.thickness_space == "world":
                vertex_array_sub = la.vec_transform(
                    positions_array_sub, wobject.world.matrix
                )
            else:  # wobject.material.thickness_space == "screen":
                xyz = la.vec_transform(
                    positions_array_sub, camera.camera_matrix @ wobject.world.matrix
                )
                vertex_array_sub = xyz[:, :2] * (0.5 * np.array(logical_size))
            # Fix up
            if has_non_finites:
                vertex_array = np.full((len(positions_array), 2), np.nan, np.float32)
                vertex_array[finites] = vertex_array_sub
            else:
                vertex_array = vertex_array_sub

        # Calculate distances
        distances = np.linalg.norm(vertex_array[1:] - vertex_array[:-1], axis=1)
        distances[~np.isfinite(distances)] = 0.0

        # Store cumulatives
        np.cumsum(distances, out=distance_array[1:])

        # Mark that the data has changed
        self.line_distance_buffer.update_range(r_offset, r_size)

    def get_bindings(self, wobject, shared):
        material = wobject.material
        geometry = wobject.geometry

        positions1 = geometry.positions

        # With vertex buffers, if a shader input is vec4, and the vbo has
        # Nx2, the z and w element will be zero. This works, because for
        # vertex buffers we provide additional information about the
        # striding of the data.
        # With storage buffers (aka SSBO) we just have some bytes that we
        # read from/write to in the shader. This is more free, but it means
        # that the data in the buffer must match with what the shader
        # expects. In addition to that, there's this thing with vec3's which
        # are padded to 16 bytes. So we either have to require our users
        # to provide Nx4 data, or read them as an array of f32.
        # Anyway, extra check here to make sure the data matches!
        if positions1.data.shape[1] != 3:
            raise ValueError(
                "For rendering (thick) lines, the geometry.positions must be Nx3."
            )

        uniform_buffer = Buffer(
            array_from_shadertype(renderer_uniform_type), force_contiguous=True
        )
        uniform_buffer.data["last_i"] = positions1.nitems - 1

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("u_renderer", "buffer/uniform", uniform_buffer),
            Binding("s_positions", rbuffer, positions1, "VERTEX"),
        ]

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] in ("vertex", "face"):
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] in ("vertex_map", "face_map"):
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )
            bindings.extend(
                self.define_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        # Need a buffer for the cumdist?
        if hasattr(self, "line_distance_buffer"):
            bindings.append(
                Binding("s_cumdist", rbuffer, self.line_distance_buffer, "VERTEX")
            )

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        # Cull backfaces so that overlapping faces are not drawn.
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def _get_n(self, positions):
        offset, size = positions.draw_range
        return offset * 6, size * 6

    def get_render_info(self, wobject, shared):
        material = wobject.material
        # Determine how many vertices are needed
        offset, size = self._get_n(wobject.geometry.positions)

        render_mask = 0
        if wobject.render_mask:
            render_mask = wobject.render_mask
        elif material.is_transparent:
            render_mask = RenderMask.transparent
        else:
            # Get what passes are needed for the color
            if self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask |= RenderMask.transparent
                else:
                    render_mask |= RenderMask.opaque
            elif self["color_mode"] in ("vertex", "face"):
                if self["color_buffer_channels"] in (2, 4):
                    render_mask |= RenderMask.all
                else:
                    render_mask |= RenderMask.opaque
            elif self["color_mode"] in ("vertex_map", "face_map"):
                if self["colormap_nchannels"] in (2, 4):
                    render_mask |= RenderMask.all
                else:
                    render_mask |= RenderMask.opaque
            else:
                raise RuntimeError(f"Unexpected color mode {self['color_mode']}")
            # Need transparency for aa
            if material.aa:
                render_mask |= RenderMask.transparent

        return {
            "indices": (size, 1, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return load_wgsl("line.wgsl")


@register_wgpu_render_function(Line, LineDebugMaterial)
class LineDebugShader(LineShader):
    def __init__(self, wobject):
        super().__init__(wobject)

        self["debug"] = True


@register_wgpu_render_function(Line, LineSegmentMaterial)
class LineSegmentShader(LineShader):
    """This shader is baded on the normal line shader, but it does not draw joins.
    Still needs 6 vertices in for nodes that have a cap on each side.
    """

    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "segment"


@register_wgpu_render_function(Line, LineArrowMaterial)
class LineArrowShader(LineShader):
    """Shader to draw arrows. This shader does not use the caps, so it could be drawn
    with less vertices, but that'd make the code more complex, so for now this is fine.
    """

    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "arrow"


# -----  shaders for thin lines


@register_wgpu_render_function(Line, LineThinMaterial)
class ThinLineShader(LineShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        self["aa"] = False  # no aa with thin lines
        if self["color_mode"] in ("face", "face_map"):
            raise RuntimeError("Face coloring not supported for thin lines.")

    def get_bindings(self, wobject, shared):
        material = wobject.material
        geometry = wobject.geometry

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
        ]

        # Per-vertex color, colormap, or a uniform color?
        if self["color_mode"] == "vertex":
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif self["color_mode"] == "vertex_map":
            bindings.append(
                Binding("s_texcoords", rbuffer, geometry.texcoords, "VERTEX")
            )
            bindings.extend(
                self.define_colormap(
                    material.map, geometry.texcoords, material.map_interpolation
                )
            )

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.line_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        material = wobject.material
        offset, size = wobject.geometry.positions.draw_range
        render_mask = wobject.render_mask
        if not render_mask:
            render_mask = RenderMask.all
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask = RenderMask.transparent
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] == "vertex":
                if self["color_buffer_channels"] in (2, 4):
                    render_mask = RenderMask.all
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] == "vertex_map":
                if self["colormap_nchannels"] in (2, 4):
                    render_mask = RenderMask.all
                else:
                    render_mask = RenderMask.opaque
        return {
            "indices": (size, 1, offset, 0),
            "render_mask": render_mask,
        }

    def get_code(self):
        return """//wgsl

        {$ include 'pygfx.std.wgsl' $}

        struct VertexInput {
            @builtin(vertex_index) index : u32,
        };

        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            let i0 = i32(in.index);

            let raw_pos = load_s_positions(i0);
            let wpos = u_wobject.world_transform * vec4<f32>(raw_pos.xyz, 1.0);
            let npos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * wpos;

            var varyings: Varyings;
            varyings.position = vec4<f32>(npos);
            varyings.world_pos = vec3<f32>(ndc_to_world_pos(npos));

            // per-vertex or per-face coloring
            $$ if color_mode == 'vertex'
                let color_index = i0;
                $$ if color_buffer_channels == 1
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
                $$ elif color_buffer_channels == 2
                    let cvalue = load_s_colors(color_index);
                    varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
                $$ elif color_buffer_channels == 3
                    varyings.color = vec4<f32>(load_s_colors(color_index), 1.0);
                $$ elif color_buffer_channels == 4
                    varyings.color = vec4<f32>(load_s_colors(color_index));
                $$ endif
            $$ endif

            // Set texture coords
            let tex_coord_index = i0;
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(tex_coord_index));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(tex_coord_index));
            $$ endif

            return varyings;
        }

        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {

            $$ if color_mode == 'vertex'
                let color = varyings.color;
            $$ elif color_mode == 'vertex_map'
                let color = sample_colormap(varyings.texcoord);
            $$ else
                let color = u_material.color;
            $$ endif

            let physical_color = srgb2physical(color.rgb);
            let opacity = color.a * u_material.opacity;
            let out_color = vec4<f32>(physical_color, opacity);

            apply_clipping_planes(varyings.world_pos);
            return get_fragment_output(varyings.position.z, out_color);
        }
        """


@register_wgpu_render_function(Line, LineThinSegmentMaterial)
class ThinLineSegmentShader(ThinLineShader):
    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.line_list,
            "cull_mode": wgpu.CullMode.none,
        }



"""
Utility functions for Pygfx.

.. currentmodule:: pygfx.utils

.. autosummary::
    :toctree: utils/
    :template: ../_templates/custom_layout.rst

    Color
    load.load_mesh
    load.load_meshes
    load.load_scene
    show.show
    show.Display
    viewport.Viewport
    text.FontManager
    enums
    cm

Transform classes
-----------------

These classes are used internally to create an intuitive API to transform
WorldObjects (`obj.world` and `obj.local`). They are listed here for
completeness and you will likely never have to instantiate them directly.

.. autosummary::
    :toctree: utils/
    :template: ../_templates/custom_layout.rst

    transform.AffineBase
    transform.AffineTransform
    transform.RecursiveTransform

"""

import os
import logging

import numpy as np

from .color import Color  # noqa: F401
from . import enums  # noqa: F401
from . import cm  # noqa: F401
from ._dirs import get_resources_dir, get_cache_dir  # noqa: F401

logger = logging.getLogger("pygfx")


def _set_log_level():
    # Set default level
    logger.setLevel(logging.WARN)
    # Set user-specified level
    level = os.getenv("PYGFX_LOG_LEVEL", "")
    if level:
        try:
            if level.isnumeric():
                logger.setLevel(int(level))
            else:
                logger.setLevel(level.upper())
        except Exception:
            logger.warn(f"Invalid pygfx log level: {level}")


_set_log_level()


def array_from_shadertype(shadertype, count=None):
    """Get a numpy array object from a dict shadertype.
    params:
        shadertype: dict
            A dict containing the shadertype.
        count: None or int
            If count is not None, array has a shape of (count, ),
            Indicates that the corresponding "Buffer" is an array of struct.
    """
    assert isinstance(shadertype, dict)
    assert count is None or count > 0

    primitives = {
        "i4": "int32",
        "u4": "uint32",
        "f4": "float32",
    }

    # Unravel the dict, turning it into a numpy array.
    # We also sort the fields so that the uniforms are properly aligned.
    # See https://www.w3.org/TR/WGSL/#structure-layout-rules
    # Note that 3x4xf4 matches a mat3x4<f32>
    array_names = []
    dtype_fields = []
    for name, format in shadertype.items():
        if format[-2:] not in primitives:
            raise RuntimeError(
                f"Values in a uniform must have a 32bit primitive type, not {format}"
            )
        primitive = primitives[format[-2:]]
        # Get shape, excluding array part
        shapestr = format[:-2].split("*")[-1]
        shape = [int(i) for i in shapestr.split("x") if i]
        align_size = shape[-1] if shape else 1  # in mat2x4 we need the 4
        if align_size == 3:  # vec3 and matnx3 are forbidden for now
            raise ValueError(
                f"Uniform format {format} forbidden for now due to alignment."
            )
        shape.reverse()  # reverse because numpy is row-major
        # Include array size
        if "*" in format:
            array_names.append(name)
            shape.insert(0, int(format.split("*")[0]))
        # Create field, include align_size for sorting
        dtype_fields.append((name, primitive, tuple(shape), align_size))

    # Sort by alignment, then strip the align_size (helper element) from the tuple
    dtype_fields.sort(key=lambda field: -field[-1])
    dtype_fields = [field[:-1] for field in dtype_fields]

    # Add meta field (zero bytes)
    # This isn't particularly pretty, but this way our metadata is attached
    # to the dtype without affecting its size.
    array_names.insert(0, "")
    array_names.append("")
    dtype_fields.append(("__".join(array_names), "uint8", (0,)))

    # Add padding: uniform buffers must align to 16 bytes.
    size = np.dtype(dtype_fields).itemsize
    n16 = int(np.ceil(size / 16))
    padding = n16 * 16 - size
    dtype_fields.append(("__padding", "uint8", (padding,)))

    # Create a scalar of this type
    if count is not None:
        uniform_data = np.zeros((count,), dtype=dtype_fields)
    else:
        uniform_data = np.zeros((), dtype=dtype_fields)

    return uniform_data


def unpack_bitfield(packed, **bit_counts):
    """Unpack values from an uint64 bitfield."""
    values = {}
    for key, bits in bit_counts.items():
        mask = 2**bits - 1
        values[key] = packed & mask
        packed = packed >> bits
    return values


def normals_from_vertices(rr, tris):
    """Efficiently compute vertex normals for a triangulated surface."""
    # This code was taken from Vispy
    # ensure highest precision for our summation/vectorization "trick"
    rr = rr[:, :3].astype(np.float64)
    tris = tris.reshape(-1, 3)
    # first, compute triangle normals
    r1 = rr[tris[:, 0], :]
    r2 = rr[tris[:, 1], :]
    r3 = rr[tris[:, 2], :]
    tri_nn = np.cross((r2 - r1), (r3 - r1))

    # Triangle normals and areas
    size = np.sqrt(np.sum(tri_nn * tri_nn, axis=1))
    size[size == 0] = 1.0  # prevent ugly divide-by-zero
    tri_nn /= size[:, np.newaxis]

    npts = len(rr)

    # the following code replaces this, but is faster (vectorized):
    #
    # for p, verts in enumerate(tris):
    #     nn[verts, :] += tri_nn[p, :]
    #
    nn = np.zeros((npts, 3))
    for verts in tris.T:  # note this only loops 3x (number of verts per tri)
        for idx in range(3):  # x, y, z
            nn[:, idx] += np.bincount(
                verts.astype(np.int32), tri_nn[:, idx], minlength=npts
            )
    size = np.sqrt(np.sum(nn * nn, axis=1))
    size[size == 0] = 1.0  # prevent ugly divide-by-zero
    nn /= size[:, np.newaxis]
    return nn.astype(np.float32)



import re


varying_types = ["f32", "vec2<f32>", "vec3<f32>", "vec4<f32>"]
varying_types = (
    varying_types
    + [t.replace("f", "i") for t in varying_types]
    + [t.replace("f", "u") for t in varying_types]
)

re_varying_getter = re.compile(r"[\s,\(\[]varyings\.(\w+)", re.UNICODE)
re_varying_setter = re.compile(r"\A\s*?varyings\.(\w+)(\.\w+)?\s*?\=")
builtin_varyings = {"position": "vec4<f32>"}


def resolve_varyings(wgsl):
    """Resolve varyings in the given wgsl:
    * Detect varyings being used.
    * Check that these are also set.
    * Remove assignments of varyings that are not used.
    * Include the Varyings struct.
    """
    assert isinstance(wgsl, str)

    # Split into lines, which is easier to process. Ensure it ends with newline in the end.
    lines = wgsl.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    while lines and not lines[-1].strip():
        lines.pop(-1)
    lines.append("")

    # Prepare dicts that map name to list-of-linenr. And a tupe dict.
    assigned_varyings = {}
    used_varyings = {}
    types = {}  # varying types

    # We try to find the function that first uses the Varyings struct.
    struct_insert_pos = None

    # Go over all lines to:
    # - find the lines where a varying is set
    # - collect the types of these varyings
    for linenr, line in enumerate(lines):
        match = re_varying_setter.match(line)
        if match:
            # Get parts
            name = match.group(1)
            attr = match.group(2)
            # Handle builtin
            if name in builtin_varyings:
                used_varyings[name] = []
                types[name] = builtin_varyings[name]
            # Find type
            type = line[match.end() :].split("(")[0].strip().replace(" ", "")
            if type not in varying_types:
                type = ""
            # Triage
            if attr:
                pass  # Not actually a type but an attribute access
            elif not type:
                raise TypeError(
                    f"Varying {name!r} assignment needs an explicit cast (of a correct type), e.g. `varying.{name} = f32(3.0);`:\n{line}"
                )
            elif name in types and type != types[name]:
                raise TypeError(
                    f"Varying {name!r} assignment does not match expected type {types[name]}:\n{line}"
                )
            else:
                types[name] = type
            # Store position
            assigned_varyings.setdefault(name, []).append(linenr)

    # Go over all lines to:
    # - collect all used varyings
    # - find where the vertex-shader starts
    in_vertex_shader = False
    current_func_linenr = 0
    for linenr, line in enumerate(lines):
        line = line.strip()
        # Detect when we enter a new function
        if line.startswith("fn "):
            current_func_linenr = linenr
            if line.startswith("fn vs_main"):
                in_vertex_shader = True
            else:
                in_vertex_shader = False
        # Remove comments (shader code has no strings that can contain slashes)
        line = line.split("//")[0]
        if "Varyings" in line and struct_insert_pos is None:
            struct_insert_pos = current_func_linenr
        # Everything we find here is a match (prepend a space to allow an easier regexp)
        for match in re_varying_getter.finditer(" " + line):
            name = match.group(1)
            this_varying_is_set_on_this_line = linenr in assigned_varyings.get(name, [])
            if this_varying_is_set_on_this_line:
                pass
            elif in_vertex_shader:
                # If varyings are used in another way than setting, in the vertex shader,
                # we should either consider them "used", or possibly break the shader if
                # the used varying is disabled. So let's just not allow it.
                raise TypeError(
                    f"Varying {name!r} is read in the vertex shader, but only writing is allowed:\n{line}"
                )
            else:
                used_varyings.setdefault(name, []).append(linenr)

    # Check if all used varyings are assigned
    for name in used_varyings:
        if name not in assigned_varyings:
            line = lines[used_varyings[name][0]]
            raise TypeError(f"Varying {name!r} is read, but not assigned:\n{line}")

    # Comment-out the varying setter if its unused elsewhere in the shader
    for name, linenrs in assigned_varyings.items():
        if name not in used_varyings:
            for linenr in linenrs:
                line = lines[linenr]
                indent = line[: len(line) - len(line.lstrip())]
                lines[linenr] = indent + "// unused: " + line[len(indent) :]
                # Deal with multiple-line assignments
                line_s = line.strip()
                while not line_s.endswith(";"):
                    linenr += 1
                    line_s = lines[linenr].strip()
                    unexpected = "fn ", "struct ", "var ", "let ", "}"
                    if line_s.startswith(unexpected) or linenr == len(lines) - 1:
                        raise TypeError(
                            f"Varying {name!r} assignment seems to be missing a semicolon:\n{line}"
                        )
                    lines[linenr] = indent + "// " + line_s

    # Build and insert the struct
    if struct_insert_pos is not None:
        # Maybe we should move up a bit
        if struct_insert_pos > 0:
            if lines[struct_insert_pos - 1].lstrip().startswith("@"):
                struct_insert_pos -= 1
        # First divide into slot-based and builtins
        used_varyings = set(used_varyings)
        used_builtins = used_varyings.intersection(builtin_varyings)
        used_slots = used_varyings.difference(used_builtins)
        used_slots = list(sorted(used_slots))
        # Build struct
        struct_lines = ["struct Varyings {"]
        for slotnr, name in enumerate(used_slots):
            struct_lines.append(f"    @location({slotnr}) {name} : {types[name]},")
        for name in sorted(used_builtins):
            struct_lines.append(f"    @builtin({name}) {name} : {types[name]},")
        struct_lines.append("};\n")
        # Apply indentation and insert
        line = lines[struct_insert_pos]
        indent = line[: len(line) - len(line.lstrip())]
        struct_lines = [indent + line for line in struct_lines]
        lines.insert(struct_insert_pos, "\n".join(struct_lines))
    else:
        assert not used_varyings, "woops, did not expect used_varyings here"

    # Return modified code
    return "\n".join(lines)


re_depth_setter = re.compile(r"\A\s*?out\.depth\s*?\=")


def resolve_depth_output(wgsl):
    """When out.depth is set (in the fragment shader), adjust the FragmentOutput
    to accept depth.
    """
    assert isinstance(wgsl, str)

    # Split into lines, which is easier to process. Ensure it ends with newline in the end.
    lines = wgsl.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    while lines and not lines[-1].strip():
        lines.pop(-1)
    lines.append("")

    # Detect whether the depth is set in the shader. We're going to assume
    # this is in the fragment shader. We check for "out.depth =".
    # Background: by default the depth is based on the geometry (set
    # by vertex shader and interpolated). It is possible for a fragment
    # shader to write the depth instead. If this is done, the GPU cannot
    # do early depth testing; the fragment shader must be run for the
    # depth to be known.
    depth_is_set = False
    struct_linenr = -1
    for linenr, line in enumerate(lines):
        if line.lstrip().startswith("struct FragmentOutput {"):
            struct_linenr = linenr
        elif re_depth_setter.match(line):
            depth_is_set = True
            if struct_linenr >= 0:
                break

    if depth_is_set:
        if struct_linenr < 0:
            raise TypeError("FragmentOutput definition not found.")
        depth_field = "    @builtin(frag_depth) depth : f32,"
        line = lines[struct_linenr]
        indent = line[: len(line) - len(line.lstrip())]
        lines.insert(struct_linenr + 1, indent + depth_field)

    return "\n".join(lines)



import pylinalg as la
from typing import List
from .. import Geometry, Line, LineSegmentMaterial, Bone, WorldObject


class SkeletonHelper(Line):
    """A helper object to assist with visualizing a Skeleton"""

    def __init__(self, wobject: WorldObject, thickness=1.0):

        bones = self._get_bones(wobject)

        positions = []
        colors = []
        for bone in bones:
            if bone.parent and isinstance(bone.parent, Bone):
                positions.append([0, 0, 0])
                positions.append([0, 0, 0])

                colors.append([0, 0, 1])
                colors.append([0, 1, 0])

        super().__init__(
            Geometry(positions=positions, colors=colors),
            LineSegmentMaterial(
                thickness=thickness, color_mode="vertex", depth_test=False
            ),
        )

        self.root = wobject
        self.bones = bones

        self.local.matrix = wobject.world.matrix

        # the helper matrix always follows the root object
        def _update_matrix(*args):
            self.local.matrix = self.root.world.matrix

        self.root.world.on_update(_update_matrix)

    def update(self):
        # TODO: we should update it automatically by some mechanism.
        # See: https://github.com/pygfx/pygfx/pull/715#issuecomment-2046493145
        """Update the helper object to match the Skeleton's bones."""
        bones = self.bones
        geometry = self.geometry
        root_matrix_world_inv = self.root.world.inverse_matrix
        positions = geometry.positions

        j = 0

        for bone in bones:
            if bone.parent and isinstance(bone.parent, Bone):
                bone_matrix = root_matrix_world_inv @ bone.world.matrix
                positions.data[j] = la.mat_decompose_translation(bone_matrix)

                parent_bone_matrix = root_matrix_world_inv @ bone.parent.world.matrix
                positions.data[j + 1] = la.mat_decompose_translation(parent_bone_matrix)

                j += 2

        positions.update_range()

    def _get_bones(self, obj) -> List[Bone]:
        # Recursively get all bones from the object and its children.
        bones = []
        if isinstance(obj, Bone):
            bones.append(obj)

        for child in obj.children:
            bones.extend(self._get_bones(child))

        return bones



import numpy as np

from ..objects._base import WorldObject
from ..utils.transform import mat_inv


class Camera(WorldObject):
    """Abstract base camera.

    Camera's are world objects and be placed in the scene, but this is not required.

    The purpose of a camera is to define the viewpoint for rendering a scene.
    This viewpoint consists of its position and orientation (in the world) and
    its projection.

    In other words, it covers the projection of world coordinates to
    normalized device coordinates (NDC), by the (inverse of) the
    camera's own world matrix and the camera's projection transform.
    The former represent the camera's position, the latter is specific
    to the type of camera.
    """

    _FORWARD_IS_MINUS_Z = True

    def __init__(self):
        super().__init__()

        self.projection_matrix = np.eye(4, dtype=float)
        self.projection_matrix_inverse = np.eye(4, dtype=float)

    def set_view_size(self, width, height):
        # In logical pixels, called by the renderer to set the viewport size
        pass

    def update_projection_matrix(self):
        raise NotImplementedError()

    def get_state(self):
        """Get the state of the camera as a dict."""
        return {}

    def set_state(self, state):
        """Set the state of the camera from a dict."""
        pass

    @property
    def view_matrix(self) -> np.ndarray:
        return self.world.inverse_matrix

    @property
    def camera_matrix(self) -> np.ndarray:
        return self.projection_matrix @ self.view_matrix


class NDCCamera(Camera):
    """A Camera operating in NDC coordinates.

    Its projection matrix is the identity transform (but its position and rotation can still be set).

    In the NDC coordinate system of wgpu (and Pygfx), x and y are in
    the range -1..1, z is in the range 0..1, and (-1, -1, 0) represents
    the bottom left corner.
    """

    def update_projection_matrix(self):
        eye = np.eye(4)
        self.projection_matrix = eye
        self.projection_matrix_inverse = eye


class ScreenCoordsCamera(Camera):
    """A Camera operating in screen coordinates.

    The depth range is the same as in NDC (0 to 1).
    """

    def __init__(self):
        super().__init__()
        self._width = 1
        self._height = 1

    def set_view_size(self, width, height):
        self._width = width
        self._height = height

    def update_projection_matrix(self):
        sx, sy, sz = 2 / self._width, 2 / self._height, 1
        dx, dy, dz = -1, -1, 0
        m = sx, 0, 0, dx, 0, sy, 0, dy, 0, 0, sz, dz, 0, 0, 0, 1
        proj_view = self.projection_matrix.ravel()
        proj_view[:] = m
        self.projection_matrix_inverse = mat_inv(self.projection_matrix)



"""Provides utilities to deal with color."""

# Possible improvements:
#
# Colorspaces
# * Support for HSLuv, a colorspace with uniform brightness https://www.hsluv.org/
#
# CSS:
# * Support "rgb(100%, 50%, 0%)"
# * Support "hsl(...)" and "hsla(...)"
#
# Other:
# * Color.lerp(color, t) linear interpolate towards other color.
# * Color.lerpHSL(color, t) same as lerp but interpolate in HSL / HSLuv space.
# * Color.lighter(factor) and Color.darker(factor)


import ctypes
import colorsys


F4 = ctypes.c_float * 4


def _float_from_css_value(v, i):
    v = v.strip()
    if v.endswith("%"):
        return float(v[:-1]) / 100
    elif i < 3:
        return float(v) / 255
    else:
        return float(v)


class Color:
    """A representation of color (in the sRGB colorspace).

    Internally the color is stored using 4 32-bit floats (rgba). It can be
    instantiated in a variety of ways. E.g. by providing the color components as
    values between 0 and 1:

        * `Color(r, g, b, a)` providing rgba values.
        * `Color(r, g, b)` providing rgb, alpha is 1.
        * `Color(gray, a)` grayscale intensity and alpha.
        * `Color(gray)` grayscale intensity.

    The above variations can also be supplied as a single tuple/list, or
    anything that:

        * `Color((r, g, b))`.

    Named colors:

        * `Color("red")` base color names.
        * `Color("cornflowerblue")` CSS color names.
        * `Color("m")` Matlab color chars.

    Hex colors:

        * `Color("#ff0000")` the common hex format.
        * `Color("#ff0000ff")` the hex format that includes alpha.
        * `Color("#ff0")` the short form hex format.
        * `Color("#ff0f")` the short form hex format that includes alpha.

    CSS color functions:

        * `Color("rgb(255, 0, 0)")`.
        * `Color("rgba(255, 0, 0, 1.0)")`.

    Parameters
    ----------
    args : tuple, int, str
        The color specification. Check the docstring of this function for
        details on available format options.

    """

    # Internally, the color is a ctypes float array
    __slots__ = ["_val"]

    def __init__(self, *args):
        if len(args) == 1:
            color = args[0]
            if isinstance(color, (int, float)):
                self._set_from_tuple(args)
            elif isinstance(color, str):
                self._set_from_str(color)
            else:
                # Assume it's an iterable,
                # may raise TypeError 'object is not iterable'
                self._set_from_tuple(color)
        else:
            self._set_from_tuple(args)

    def __repr__(self):
        # A precision of 4 decimals, i.e. 10001 possible values for each color.
        # We truncate zeros, but make sure the value does not end with a dot.
        f = lambda v: f"{v:0.4f}".rstrip("0").ljust(3, "0")  # noqa: stfu
        return f"Color({f(self.r)}, {f(self.g)}, {f(self.b)}, {f(self.a)})"

    @property
    def __array_interface__(self):
        # Numpy can wrap our memory in an array without copying
        readonly = True
        ptr = ctypes.addressof(self._val)
        x = dict(version=3, shape=(4,), typestr="<f4", data=(ptr, readonly))
        return x

    def __len__(self):
        return 4

    def __getitem__(self, index):
        return self._val[index]

    def __iter__(self):
        return self.rgba.__iter__()

    def __eq__(self, other):
        if not isinstance(other, Color):
            other = Color(other)
        return all(self._val[i] == other._val[i] for i in range(4))

    def __add__(self, other):
        return Color(
            self.r + other.r,
            self.g + other.g,
            self.b + other.b,
            self.a,
        )

    def __mul__(self, factor):
        if not isinstance(factor, (float, int)):
            raise TypeError("Can only multiple a color with a scalar.")
        return Color(
            self.r * factor,
            self.g * factor,
            self.b * factor,
            self.a,
        )

    def __truediv__(self, factor):
        if not isinstance(factor, (float, int)):
            raise TypeError("Can only multiple a color with a scalar.")
        return self.__mul__(1 / factor)

    def _set_from_rgba(self, r, g, b, a):
        a = max(0.0, min(1.0, float(a)))
        self._val = F4(float(r), float(g), float(b), a)

    def _set_from_tuple(self, color):
        color = tuple(float(c) for c in color)
        if len(color) == 4:
            self._set_from_rgba(*color)
        elif len(color) == 3:
            self._set_from_rgba(*color, 1)
        elif len(color) == 2:
            self._set_from_rgba(color[0], color[0], color[0], color[1])
        elif len(color) == 1:
            self._set_from_rgba(color[0], color[0], color[0], 1)
        else:
            raise ValueError(f"Cannot parse color tuple with {len(color)} values")

    def _set_from_str(self, color):
        color = color.lower()
        if color.startswith("#"):
            # A hex number
            if len(color) == 7:  # #rrggbb
                self._set_from_rgba(
                    int(color[1:3], 16) / 255,
                    int(color[3:5], 16) / 255,
                    int(color[5:7], 16) / 255,
                    1,
                )
            elif len(color) == 4:  # #rgb
                self._set_from_rgba(
                    int(color[1], 16) / 15,
                    int(color[2], 16) / 15,
                    int(color[3], 16) / 15,
                    1,
                )
            elif len(color) == 9:  # #rrggbbaa
                self._set_from_rgba(
                    int(color[1:3], 16) / 255,
                    int(color[3:5], 16) / 255,
                    int(color[5:7], 16) / 255,
                    int(color[7:9], 16) / 255,
                )
            elif len(color) == 5:  # #rgba
                self._set_from_rgba(
                    int(color[1], 16) / 15,
                    int(color[2], 16) / 15,
                    int(color[3], 16) / 15,
                    int(color[4], 16) / 15,
                )
            else:
                raise ValueError(
                    f"Expecting 4, 5, 7, or 9 chars in a hex number, got {len(color)}."
                )
        elif color.startswith(("rgb(", "rgba(")):
            # A CSS color 'function'
            parts = color.split("(")[1].split(")")[0].split(",")
            parts = [_float_from_css_value(p, i) for i, p in enumerate(parts)]
            if len(parts) == 3:
                self._set_from_rgba(parts[0], parts[1], parts[2], 1)
            elif len(parts) == 4:
                self._set_from_rgba(parts[0], parts[1], parts[2], parts[3])
            else:
                raise ValueError(
                    f"CSS color {color.split('(')[0]}(..) must have 3 or 4 elements, not {len(parts)} "
                )
        else:
            # Maybe a named color
            try:
                color_int = NAMED_COLORS[color.lower()]
            except KeyError:
                raise ValueError(f"Unknown color: '{color}'") from None
            else:
                self._set_from_str(color_int)

    @property
    def rgba(self):
        """The RGBA tuple (values between 0 and 1)."""
        return self._val[0], self._val[1], self._val[2], self._val[3]

    @property
    def rgb(self):
        """The RGB tuple (values between 0 and 1)."""
        return self._val[0], self._val[1], self._val[2]

    @property
    def r(self):
        """The red value."""
        return self._val[0]

    @property
    def g(self):
        """The green value."""
        return self._val[1]

    @property
    def b(self):
        """The blue value."""
        return self._val[2]

    @property
    def a(self):
        """The alpha (transparency) value, between 0 and 1."""
        return self._val[3]

    @property
    def gray(self):
        """Return the grayscale intensity."""
        # Calculate with the weighted (a.k.a. luminosity) method, same as Matlab
        r, g, b = self.rgb
        return 0.2989 * r + 0.5870 * g + 0.1140 * b

    @property
    def hex(self):
        """The CSS hex string, e.g. "#00ff00". The alpha channel is ignored.
        Values are clipped to 00 an ff.
        """
        c = self.clip()
        r = int(c.r * 255 + 0.5)
        b = int(c.b * 255 + 0.5)
        g = int(c.g * 255 + 0.5)
        i = (r << 16) + (g << 8) + b
        return "#" + hex(i)[2:].rjust(6, "0")

    @property
    def hexa(self):
        """The hex string including alpha, e.g. "#00ff00ff".
        Values are clipped to 00 an ff.
        """
        c = self.clip()
        r = int(c.r * 255 + 0.5)
        b = int(c.b * 255 + 0.5)
        g = int(c.g * 255 + 0.5)
        a = int(c.a * 255 + 0.5)
        i = (r << 24) + (g << 16) + (b << 8) + a
        return "#" + hex(i)[2:].rjust(8, "0")

    @property
    def css(self):
        """The CSS color string, e.g. "rgba(0,255,0,0.5)"."""
        r, g, b, a = self.rgba
        if a == 1:
            return f"rgb({int(255*r+0.5)},{int(255*g+0.5)},{int(255*b+0.5)})"
        else:
            return f"rgba({int(255*r+0.5)},{int(255*g+0.5)},{int(255*b+0.5)},{a:0.3f})"

    def clip(self):
        """Return a new Color with the values clipped between 0 and 1."""
        return Color(max(0.0, min(1.0, x)) for x in self.rgba)

    @classmethod
    def from_physical(cls, r, g, b, a=1):
        """Create a Color object from a color in the physical colorspace.

        With the physical colorspace we mean what is sometimes called
        Linear-sRGB. It has the same gamut as sRGB, but where sRGB is
        linear w.r.t. human perception, Linear-sRGB is linear w.r.t.
        lumen and photon counts. Calculations on colors in pygfx's shaders
        are done in the physical colorspace.
        """
        return Color(_physical2srgb(r), _physical2srgb(g), _physical2srgb(b), a)

    def to_physical(self):
        """Get the color represented in the physical colorspace, as 3 floats."""
        return _srgb2physical(self.r), _srgb2physical(self.g), _srgb2physical(self.b)

    @classmethod
    def from_hsv(cls, hue, saturation, value):
        """Create a Color object from an a color in the HSV (a.k.a. HSB) colorspace.

        HSV stands for hue, saturation, value (aka brightness). The hue
        component indicates the color tone. Values go from red (0) to
        green (0.333) to blue (0.666) and back to red (1). The
        satutation indicates vividness, with 0 meaning gray and 1
        meaning the primary color. The value/brightness indicates goes
        from 0 (black) to 1 (white).
        """
        return Color(colorsys.hsv_to_rgb(hue, saturation, value))

    def to_hsv(self):
        """Get the color represented in the HSV colorspace, as 3 floats."""
        return colorsys.rgb_to_hsv(*self.rgb)

    @classmethod
    def from_hsl(cls, hue, saturation, lightness):
        """Create a Color object from an a color in the HSL colorspace.

        The HSL colorspace is similar to the HSV colorspace, except the
        "value" is replaced with "lightness". This lightness scales the
        color differently, e.g. a lightness of 1 always represents full
        white.
        """
        return Color(colorsys.hls_to_rgb(hue, lightness, saturation))

    def to_hsl(self):
        """Get the color represented in the HSL colorspace, as 3 floats."""
        hue, lightness, saturation = colorsys.rgb_to_hls(*self.rgb)
        return hue, saturation, lightness


def _srgb2physical(c):
    # The simplified version has a maximum error less than 1%, but that's still
    # two steps in the range 0..255.
    # return c ** 2.2
    return c / 12.92 if c <= 0.04045 else ((c + 0.055) / 1.055) ** 2.4


def _physical2srgb(c):
    # return c ** (1 / 2.2)
    return c * 12.92 if c <= 0.0031308 else c ** (1 / 2.4) * 1.055 - 0.055


NAMED_COLORS = {
    # CSS Level 1
    "black": "#000000",
    "silver": "#C0C0C0",
    "gray": "#808080",
    "white": "#FFFFFF",
    "maroon": "#800000",
    "red": "#FF0000",
    "purple": "#800080",
    "fuchsia": "#FF00FF",
    "green": "#008000",
    "lime": "#00FF00",
    "olive": "#808000",
    "yellow": "#FFFF00",
    "navy": "#000080",
    "blue": "#0000FF",
    "teal": "#008080",
    "aqua": "#00FFFF",
    # CSS Level 2
    "orange": "#FFA500",
    # CSS Color Module Level 3
    "aliceblue": "#F0F8FF",
    "antiquewhite": "#FAEBD7",
    "aquamarine": "#7FFFD4",
    "azure": "#F0FFFF",
    "beige": "#F5F5DC",
    "bisque": "#FFE4C4",
    "blanchedalmond": "#FFEBCD",
    "blueviolet": "#8A2BE2",
    "brown": "#A52A2A",
    "burlywood": "#DEB887",
    "cadetblue": "#5F9EA0",
    "chartreuse": "#7FFF00",
    "chocolate": "#D2691E",
    "coral": "#FF7F50",
    "cornflowerblue": "#6495ED",
    "cornsilk": "#FFF8DC",
    "crimson": "#DC143C",
    "cyan": "#00FFFF",
    "aqua": "#00FFFF",
    "darkblue": "#00008B",
    "darkcyan": "#008B8B",
    "darkgoldenrod": "#B8860B",
    "darkgray": "#A9A9A9",
    "darkgreen": "#006400",
    "darkgrey": "#A9A9A9",
    "darkkhaki": "#BDB76B",
    "darkmagenta": "#8B008B",
    "darkolivegreen": "#556B2F",
    "darkorange": "#FF8C00",
    "darkorchid": "#9932CC",
    "darkred": "#8B0000",
    "darksalmon": "#E9967A",
    "darkseagreen": "#8FBC8F",
    "darkslateblue": "#483D8B",
    "darkslategray": "#2F4F4F",
    "darkslategrey": "#2F4F4F",
    "darkturquoise": "#00CED1",
    "darkviolet": "#9400D3",
    "deeppink": "#FF1493",
    "deepskyblue": "#00BFFF",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1E90FF",
    "firebrick": "#B22222",
    "floralwhite": "#FFFAF0",
    "forestgreen": "#228B22",
    "gainsboro": "#DCDCDC",
    "ghostwhite": "#F8F8FF",
    "gold": "#FFD700",
    "goldenrod": "#DAA520",
    "greenyellow": "#ADFF2F",
    "grey": "#808080",
    "honeydew": "#F0FFF0",
    "hotpink": "#FF69B4",
    "indianred": "#CD5C5C",
    "indigo": "#4B0082",
    "ivory": "#FFFFF0",
    "khaki": "#F0E68C",
    "lavender": "#E6E6FA",
    "lavenderblush": "#FFF0F5",
    "lawngreen": "#7CFC00",
    "lemonchiffon": "#FFFACD",
    "lightblue": "#ADD8E6",
    "lightcoral": "#F08080",
    "lightcyan": "#E0FFFF",
    "lightgoldenrodyellow": "#FAFAD2",
    "lightgray": "#D3D3D3",
    "lightgreen": "#90EE90",
    "lightgrey": "#D3D3D3",
    "lightpink": "#FFB6C1",
    "lightsalmon": "#FFA07A",
    "lightseagreen": "#20B2AA",
    "lightskyblue": "#87CEFA",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#B0C4DE",
    "lightyellow": "#FFFFE0",
    "limegreen": "#32CD32",
    "linen": "#FAF0E6",
    "magenta": "#FF00FF",
    "fuchsia": "#FF00FF",
    "mediumaquamarine": "#66CDAA",
    "mediumblue": "#0000CD",
    "mediumorchid": "#BA55D3",
    "mediumpurple": "#9370DB",
    "mediumseagreen": "#3CB371",
    "mediumslateblue": "#7B68EE",
    "mediumspringgreen": "#00FA9A",
    "mediumturquoise": "#48D1CC",
    "mediumvioletred": "#C71585",
    "midnightblue": "#191970",
    "mintcream": "#F5FFFA",
    "mistyrose": "#FFE4E1",
    "moccasin": "#FFE4B5",
    "navajowhite": "#FFDEAD",
    "oldlace": "#FDF5E6",
    "olivedrab": "#6B8E23",
    "orangered": "#FF4500",
    "orchid": "#DA70D6",
    "palegoldenrod": "#EEE8AA",
    "palegreen": "#98FB98",
    "paleturquoise": "#AFEEEE",
    "palevioletred": "#DB7093",
    "papayawhip": "#FFEFD5",
    "peachpuff": "#FFDAB9",
    "peru": "#CD853F",
    "pink": "#FFC0CB",
    "plum": "#DDA0DD",
    "powderblue": "#B0E0E6",
    "rosybrown": "#BC8F8F",
    "royalblue": "#4169E1",
    "saddlebrown": "#8B4513",
    "salmon": "#FA8072",
    "sandybrown": "#F4A460",
    "seagreen": "#2E8B57",
    "seashell": "#FFF5EE",
    "sienna": "#A0522D",
    "skyblue": "#87CEEB",
    "slateblue": "#6A5ACD",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#FFFAFA",
    "springgreen": "#00FF7F",
    "steelblue": "#4682B4",
    "tan": "#D2B48C",
    "thistle": "#D8BFD8",
    "tomato": "#FF6347",
    "turquoise": "#40E0D0",
    "violet": "#EE82EE",
    "wheat": "#F5DEB3",
    "whitesmoke": "#F5F5F5",
    "yellowgreen": "#9ACD32",
    # CSS Color Module Level 4
    "rebeccapurple": "#663399",
    # Matlab / Matplotlib
    "b": "#0000FF",
    "g": "#00FF00",
    "r": "#FF0000",
    "c": "#00FFFF",
    "m": "#FF00FF",
    "y": "#FFFF00",
    "k": "#000000",
    "w": "#FFFFFF",
}



"""
Utilities to load gltf/glb files, completely compatible with the glTF 2.0 specification.

References:
https://raw.githubusercontent.com/KhronosGroup/glTF/main/specification/2.0/figures/gltfOverview-2.0.0d.png
https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html

Experimental, not yet fully implemented.
"""

import pygfx as gfx
import numpy as np
import pylinalg as la

from importlib.util import find_spec
from functools import lru_cache


def load_gltf(path, quiet=False):
    """
    Load a gltf file and return the content.

    This function requires the gltflib library.

    Parameters:
    ----------
    path : str
        The path to the gltf file.
    quiet : bool
        Whether to suppress the warning messages.

    Returns:
    ----------
    gltf : object
        The gltf object which contains the following attributes:
        * `scenes`: [gfx.Group]
        * `scene`: gfx.Group or None
        * `cameras`: [gfx.Camera] or None
        * `animations`: [gfx.Animation] or None
    """
    return _GLTF(path, quiet).load()


def load_gltf_mesh(path, materials=True, quiet=False):
    """
    Load meshes from a gltf file, without skeletons, and no transformations applied.

    This function requires the gltflib library.

    Parameters:
    ----------
    path : str
        The path to the gltf file.
    materials : bool
        Whether to load materials.
    quiet : bool
        Whether to suppress the warning messages.

    Returns:
    ----------
    meshes : list
        A list of pygfx.Meshes.
    """
    return _GLTF(path, quiet).load_mesh(materials=materials)


class _GLTF:
    ACCESSOR_TYPE_SIZE = {
        "SCALAR": 1,
        "VEC2": 2,
        "VEC3": 3,
        "VEC4": 4,
        "MAT2": 4,
        "MAT3": 9,
        "MAT4": 16,
    }

    COMPONENT_TYPE = {
        5120: np.int8,
        5121: np.uint8,
        5122: np.int16,
        5123: np.uint16,
        5125: np.uint32,
        5126: np.float32,
    }

    ATTRIBUTE_NAME = {
        "POSITION": "positions",
        "NORMAL": "normals",
        "TANGENT": "tangents",
        "TEXCOORD_0": "texcoords",
        "TEXCOORD_1": "texcoords1",
        "COLOR_0": "colors",
        "JOINTS_0": "skin_indices",
        "WEIGHTS_0": "skin_weights",
    }

    SUPPORTED_EXTENSIONS = ["KHR_mesh_quantization"]

    def __init__(self, path, quiet=False):
        self._path = path
        self.scene = None
        self.scenes = []
        self.cameras = None
        self.animations = None

        self.__inner_load(quiet)

    def load(self):
        """Load the whole gltf file, including meshes, skeletons, cameras, and animations."""

        self.scenes = self._load_scenes()
        if self._gltf.model.scene is not None:
            self.scene = self.scenes[self._gltf.model.scene]
        if self._gltf.model.animations is not None:
            self.animations = self._load_animations()

        # TODO:
        # self.cameras
        return self

    def load_mesh(self, materials=True):
        """Only load meshes from a gltf file, without skeletons, and no transformations applied."""

        meshes = []
        for gltf_mesh in self._gltf.model.meshes:
            mesh = self._load_gltf_mesh_by_info(gltf_mesh, load_material=materials)
            meshes.extend(mesh)
        return meshes

    def __inner_load(self, quiet=False):
        if not find_spec("gltflib"):
            raise ImportError(
                "The `gltflib` library is required to load gltf scene: pip install gltflib"
            )
        import gltflib  # noqa

        path = self._path
        self._gltf = gltflib.GLTF.load(path, load_file_resources=True)

        if not quiet:
            extensions_required = self._gltf.model.extensionsRequired or []

            unsupported_extensions_required = set(extensions_required) - set(
                self.SUPPORTED_EXTENSIONS
            )

            if unsupported_extensions_required:
                gfx.utils.logger.warning(
                    f"This GLTF required extensions: {unsupported_extensions_required}, which are not supported yet."
                )
                # rasise or ignore?
                # raise NotImplementedError(f"This GLTF required extensions: {extensions_required}, which are not supported yet.")

            extensions_used = self._gltf.model.extensionsUsed or []

            unsupported_extensions_used = set(extensions_used) - set(
                self.SUPPORTED_EXTENSIONS
            )
            if unsupported_extensions_used:
                gfx.utils.logger.warning(
                    f"This GLTF used extensions: {unsupported_extensions_used}, which are not supported yet, so the display may not be so correct."
                )

        # bind the actual data to the buffers
        for buffer in self._gltf.model.buffers:
            buffer.data = self._get_resource_by_uri(buffer.uri).data

        # mark the node types
        self._node_marks = self._mark_nodes()

    @lru_cache(maxsize=None)
    def _get_resource_by_uri(self, uri):
        for resource in self._gltf.resources:
            if resource.uri == uri:
                return resource
        raise ValueError(f"Buffer data not found for buffer {uri}")

    def _mark_nodes(self):
        gltf = self._gltf
        node_marks = [None] * len(gltf.model.nodes)

        # Nothing in the node definition indicates whether it is a Bone.
        # Use the skins' joint references to mark bones.
        if gltf.model.skins:
            for skin in gltf.model.skins:
                for joint in skin.joints:
                    node_marks[joint] = "Bone"

        # Mark cameras
        if gltf.model.cameras:
            for camera in gltf.model.cameras:
                node_marks[camera.node] = "Camera"

        # Meshes are marked when they are loaded
        # Maybe mark lights and other special nodes here
        return node_marks

    def _load_scenes(self):
        gltf = self._gltf
        scenes = []
        for scene in gltf.model.scenes:
            scene_obj = gfx.Group()
            scene_obj.name = scene.name
            scenes.append(scene_obj)

            for node in scene.nodes:
                node_obj = self._load_node(node)
                scene_obj.add(node_obj)

        return scenes

    @lru_cache(maxsize=None)
    def _load_node(self, node_index):
        gltf = self._gltf
        node_marks = self._node_marks

        node = gltf.model.nodes[node_index]

        if node.matrix is not None:
            matrix = np.array(node.matrix).reshape(4, 4).T
        else:
            translation = node.translation or [0, 0, 0]
            rotation = node.rotation or [0, 0, 0, 1]
            scale = node.scale or [1, 1, 1]
            matrix = la.mat_compose(translation, rotation, scale)

        node_mark = node_marks[node_index]

        if node_mark == "Bone":
            node_obj = gfx.Bone()
        elif node_mark == "Camera":
            # TODO: implement camera loading
            # node_obj = gfx.Camera()
            pass
        elif node.mesh is not None:  # Mesh or SkinnedMesh
            meshes = self._load_gltf_mesh(node.mesh, node.skin)
            if len(meshes) == 1:
                node_obj = meshes[0]
            else:
                node_obj = gfx.Group()
                for mesh in meshes:
                    node_obj.add(mesh)
        else:
            node_obj = gfx.WorldObject()

        node_obj.local.matrix = matrix
        node_obj.name = node.name

        if node.children:
            for child in node.children:
                child_obj = self._load_node(child)
                node_obj.add(child_obj)

        return node_obj

    def _load_gltf_mesh_by_info(self, mesh, skin_index=None, load_material=True):
        meshes = []
        for primitive in mesh.primitives:
            geometry = self._load_gltf_geometry(primitive)
            primitive_mode = primitive.mode

            if primitive_mode is None:
                primitive_mode = 4  # default to triangles

            if primitive_mode == 0:
                material = gfx.PointsMaterial()
                gfx_mesh = gfx.Points(geometry, material)

            elif primitive_mode in (1, 2, 3):
                material = gfx.LineSegmentMaterial()
                gfx_mesh = gfx.Line(geometry, material)

            elif primitive_mode in (4, 5):
                if load_material and primitive.material is not None:
                    material = self._load_gltf_material(primitive.material)
                else:
                    material = gfx.MeshBasicMaterial()

                if skin_index is not None:
                    gfx_mesh = gfx.SkinnedMesh(geometry, material)
                    skeleton = self._load_skins(skin_index)
                    gfx_mesh.bind(skeleton, np.identity(4))
                else:
                    gfx_mesh = gfx.Mesh(geometry, material)
            else:
                raise ValueError(f"Unsupported primitive mode: {primitive.mode}")

            if (
                getattr(geometry, "morph_positions", None)
                or getattr(geometry, "morph_normals", None)
                or getattr(geometry, "morph_colors", None)
            ):
                self.update_morph_target(gfx_mesh, mesh)

            meshes.append(gfx_mesh)

        return meshes

    def update_morph_target(self, gfx_mesh, mesh_def):
        if mesh_def.weights:
            gfx_mesh.morph_target_influences = mesh_def.weights

        if mesh_def.extras and mesh_def.extras.get("targetNames", None):
            gfx_mesh.morph_target_names.extend(mesh_def.extras["targetNames"])

    @lru_cache(maxsize=None)
    def _load_gltf_mesh(self, mesh_index, skin_index=None, load_material=True):
        mesh = self._gltf.model.meshes[mesh_index]
        return self._load_gltf_mesh_by_info(mesh, skin_index, load_material)

    @lru_cache(maxsize=None)
    def _load_gltf_material(self, material_index):
        material = self._gltf.model.materials[material_index]
        pbr_metallic_roughness = material.pbrMetallicRoughness

        if pbr_metallic_roughness is not None:
            gfx_material = gfx.MeshStandardMaterial()

            if pbr_metallic_roughness.baseColorTexture is not None:
                gfx_material.map = self._load_gltf_texture(
                    pbr_metallic_roughness.baseColorTexture
                )

            if pbr_metallic_roughness.metallicRoughnessTexture is not None:
                metallic_roughness_map = self._load_gltf_texture(
                    pbr_metallic_roughness.metallicRoughnessTexture
                )
                gfx_material.roughness_map = metallic_roughness_map
                gfx_material.metalness_map = metallic_roughness_map
                gfx_material.roughness = 1.0
                gfx_material.metalness = 1.0

            if pbr_metallic_roughness.roughnessFactor is not None:
                gfx_material.roughness = pbr_metallic_roughness.roughnessFactor

            if pbr_metallic_roughness.metallicFactor is not None:
                gfx_material.metalness = pbr_metallic_roughness.metallicFactor

        else:
            gfx_material = gfx.MeshPhongMaterial()

        if material.normalTexture is not None:
            gfx_material.normal_map = self._load_gltf_texture(material.normalTexture)
            scale_factor = material.normalTexture.scale
            if scale_factor is None:
                scale_factor = 1.0

            # pygfx now assume the normal map is in tangent space, so we need to flip the y-axis.
            # See: https://github.com/KhronosGroup/glTF-Sample-Assets/tree/main/Models/NormalTangentTest#problem-flipped-y-axis-or-flipped-green-channel
            # TODO: support object space normal map, and flip the y-axis when only in tangent space.(check if the tangent attribute in the mesh primitive)
            gfx_material.normal_scale = (scale_factor, -scale_factor)

        if material.occlusionTexture is not None:
            gfx_material.ao_map = self._load_gltf_texture(material.occlusionTexture)

        if material.emissiveFactor is not None:
            gfx_material.emissive = gfx.Color(*material.emissiveFactor)

        if material.emissiveTexture is not None:
            gfx_material.emissive_map = self._load_gltf_texture(
                material.emissiveTexture
            )

        return gfx_material

    def _load_gltf_texture(self, texture_info):
        texture_index = texture_info.index
        texture = self._load_gltf_texture_resource(texture_index)
        # uv_channel = texture_info.texCoord
        # TODO: use uv_channel when pygfx supports it
        return texture

    @lru_cache(maxsize=None)
    def _load_gltf_texture_resource(self, texture_index):
        texture_desc = self._gltf.model.textures[texture_index]
        source = texture_desc.source
        image = self._load_image(source)
        texture = gfx.Texture(image, dim=2)

        sampler = texture_desc.sampler
        sampler = self._load_gltf_sampler(sampler)
        # pygfx not support set texture sampler info now
        # TODO: implement this after pygfx support texture custom sampler
        return texture

    @lru_cache(maxsize=None)
    def _load_gltf_sampler(self, sampler_index):
        sampler = self._gltf.model.samplers[sampler_index]
        # Sampler( magFilter=9729, minFilter=9987, wrapS=None, wrapT=None)
        return sampler

    @lru_cache(maxsize=None)
    def _load_image(self, image_index):
        image_info = self._gltf.model.images[image_index]
        import imageio.v3 as iio

        if image_info.bufferView is not None:
            image_data = self._get_buffer_memory_view(image_info.bufferView)

        elif image_info.uri is not None:
            resource = self._get_resource_by_uri(image_info.uri)
            image_data = resource.data
        else:
            raise ValueError("No image data found")

        # need consider mimeType?
        image = iio.imread(image_data)
        return image

    def _load_gltf_geometry(self, primitive):
        indices_accessor = primitive.indices

        geometry_args = {}

        for attr, accessor_index in primitive.attributes.__dict__.items():
            if accessor_index is not None:
                geometry_attr = self.ATTRIBUTE_NAME[attr]
                data = self._load_accessor(accessor_index)

                # pygfx not support int attributes now, so we need to convert them to float.
                if geometry_attr in (
                    "positions",
                    "normals",
                    "tangents",
                    "texcoords",
                    "texcoords1",
                ):
                    data = data.astype(np.float32, copy=False)

                geometry_args[geometry_attr] = data

        if indices_accessor is not None:
            indices = self._load_accessor(indices_accessor).reshape(-1, 3)
        else:
            # TODO: For now, pygfx not support non-indexed geometry, so we need to generate indices for them.
            # Remove this after pygfx support non-indexed geometry.
            indices = np.arange(
                len(geometry_args["positions"]) * 3, dtype=np.int32
            ).reshape((-1, 3))

        geometry_args["indices"] = indices

        geometry = gfx.Geometry(**geometry_args)

        if primitive.targets:
            for target in primitive.targets:
                # print(target)
                for attr, accessor_index in target.__dict__.items():
                    if accessor_index is not None:
                        target_attr = f"morph_{self.ATTRIBUTE_NAME[attr]}"
                        data = self._load_accessor(accessor_index).astype(
                            np.float32, copy=False
                        )

                        morph_attr = getattr(geometry, target_attr, None)
                        if morph_attr is None:
                            morph_attr = []
                            setattr(geometry, target_attr, morph_attr)

                        morph_attr.append(data)

            geometry.morph_targets_relative = True

        return geometry

    @lru_cache(maxsize=None)
    def _get_buffer_memory_view(self, buffer_view_index):
        gltf = self._gltf
        buffer_view = gltf.model.bufferViews[buffer_view_index]
        buffer = gltf.model.buffers[buffer_view.buffer]
        m = memoryview(buffer.data)
        view = m[
            buffer_view.byteOffset : buffer_view.byteOffset + buffer_view.byteLength
        ]
        return view

    @lru_cache(maxsize=None)
    def _load_accessor(self, accessor_index):
        gltf = self._gltf
        accessor = gltf.model.accessors[accessor_index]

        buffer_view = gltf.model.bufferViews[accessor.bufferView]
        view = self._get_buffer_memory_view(accessor.bufferView)

        accessor_type = accessor.type
        accessor_component_type = accessor.componentType
        accessor_count = accessor.count
        accessor_dtype = np.dtype(self.COMPONENT_TYPE[accessor_component_type])
        accessor_offset = accessor.byteOffset or 0
        accessor_type_size = self.ACCESSOR_TYPE_SIZE[accessor_type]

        if buffer_view.byteStride is not None:
            # It's a interleaved buffer
            # pygfx not support interleaved buffer now, so we pick out the data we need from the interleaved buffer.
            # TODO: optimize this after pygfx support interleaved buffer.
            ar = np.lib.stride_tricks.as_strided(
                view[accessor_offset:],
                shape=(accessor_count, accessor_type_size * accessor_dtype.itemsize),
                strides=(buffer_view.byteStride, 1),
            )
            ar = np.frombuffer(np.ascontiguousarray(ar), dtype=accessor_dtype).reshape(
                accessor_count, accessor_type_size
            )
        else:
            ar = np.frombuffer(
                view,
                dtype=accessor_dtype,
                offset=accessor_offset,
                count=accessor_count * accessor_type_size,
            )
            if accessor_type_size > 1:
                ar = ar.reshape(accessor_count, accessor_type_size)

        if accessor.normalized:
            # KHR_mesh_quantization
            # https://github.com/KhronosGroup/glTF/tree/main/extensions/2.0/Khronos/KHR_mesh_quantization
            assert accessor_dtype.kind == "i" or accessor_dtype.kind == "u"
            ar = ar.astype(np.float32, copy=False) / np.iinfo(accessor_dtype).max

        # pygfx not support int8, int16, uint8, uint16 now
        if ar.dtype == np.uint8 or ar.dtype == np.uint16:
            ar = ar.astype(np.uint32, copy=False)
        if ar.dtype == np.int8 or ar.dtype == np.int16:
            ar = ar.astype(np.int32, copy=False)

        return ar

    @lru_cache(maxsize=None)
    def _load_skins(self, skin_index):
        skin = self._gltf.model.skins[skin_index]
        bones = [self._load_node(index) for index in skin.joints]
        inverse_bind_matrices = self._load_accessor(skin.inverseBindMatrices)

        bone_inverses = []
        for matrices in inverse_bind_matrices:
            bone_inverse = np.array(matrices).reshape(4, 4).T
            bone_inverses.append(bone_inverse)

        skeleton = gfx.Skeleton(bones, bone_inverses)
        return skeleton

    def _load_animation(self, animation_info):
        channels = animation_info.channels
        samplers = animation_info.samplers

        duration = 0

        key_frame_tracks = []

        for channel in channels:
            target = channel.target
            sampler = samplers[channel.sampler]

            target_node = self._load_node(target.node)
            name = target_node.name
            target_property = target.path
            interpolation = sampler.interpolation
            times = self._load_accessor(sampler.input)
            if times[-1] > duration:
                duration = times[-1]
            values = self._load_accessor(sampler.output)

            key_frame_tracks.append(
                {
                    "name": name,
                    "target": target_node,
                    "property": target_property,
                    "interpolation": interpolation,
                    "times": times,
                    "values": values,
                }
            )

        action_clip = {
            "name": animation_info.name,
            "duration": duration,
            "tracks": key_frame_tracks,
        }

        return action_clip

    def _load_animations(self):
        gltf = self._gltf
        animations = []
        for animation in gltf.model.animations:
            action_clip = self._load_animation(animation)
            animations.append(action_clip)
        return animations


def print_tree(obj, show_pos=False, show_rot=False, show_scale=False):
    """
    Print the tree structure of the scene, including the optional position, rotation, and scale of each object.

    Parameters:
    ----------
    obj : gfx.WorldObject
        The root object.
    show_pos : bool
        Whether to show the position of each object.
    show_rot : bool
        Whether to show the rotation of each object.
    show_scale : bool
        Whether to show the scale of each object.
    """

    def _print_tree(obj: gfx.WorldObject, level=0):

        name = "- " * level + f"{obj.__class__.__name__}[{obj.name}]"
        if show_pos:
            name += f"\n{'  ' * level}|- pos: {obj.local.position}"
        if show_rot:
            name += f"\n{'  ' * level}|- rot: {obj.local.rotation}"
        if show_scale:
            name += f"\n{'  ' * level}|- scale: {obj.local.scale}"

        print(name)

        for child in obj.children:
            _print_tree(child, level=level + 1)

    _print_tree(obj)



from ._base import Material
from ..utils import unpack_bitfield, Color


class TextMaterial(Material):
    """Basic text material.

    Parameters
    ----------
    color : Color
        The color of the text.
    outline_color : Color
        The color of the outline of the text.
    outline_thickness : int
        A value indicating the relative width of the outline. Valid values are
        between 0.0 and 0.5.
    weight_offset : int
        A value representing an offset to the font weight. Font weights are in
        the range 100-900, so this value should be in the same order of
        magnitude. Can be negative to make text thinner. Default zero.
    aa : bool
        If True, use anti-aliasing while rendering glyphs. Aliasing gives
        prettier results, but may affect performance for very large texts.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    Notes
    -----
    One use-case for weight_offset is to make dark text on a light background 50
    units thicker, to counter the psychological effect that such text *looks*
    thinner than a light text on a dark background.

    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        weight_offset="f4",
        outline_thickness="f4",
        outline_color="4xf4",
    )

    def __init__(
        self,
        color="#fff",
        *,
        outline_color="#000",
        outline_thickness=0,
        weight_offset=0,
        aa=True,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.color = color
        self.outline_color = outline_color
        self.outline_thickness = outline_thickness
        self.weight_offset = weight_offset
        self.aa = aa

    def _wgpu_get_pick_info(self, pick_value):
        # Note that the glyph index is not necessarily the same as the
        # char index. It would not be worth the effort to let the shader produce
        # the char index, but I think we could make it possible to look
        # up the char index from the glyph index via the geometry.
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, x=9, y=9)
        return {
            "glyph_index": values["index"],
            "point_coord": (values["x"] / 511.0, values["y"] / 511.0),
        }

    @property
    def aa(self):
        """Whether or not the glyphs should be anti-aliased. Aliasing
        gives prettier results, but may affect performance for very large
        texts. Default True.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def color(self):
        """The color of the text."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_full()
        self._store.color_is_transparent = color.a < 1

    @property
    def color_is_transparent(self):
        """Whether the color is (semi) transparent (i.e. not fully opaque)."""
        return self._store.color_is_transparent

    @property
    def outline_thickness(self):
        """A value indicating the relative width of the outline. Valid
        values are between 0.0 and 0.5. Default 0 (no outline).
        """
        return float(self.uniform_buffer.data["outline_thickness"])

    @outline_thickness.setter
    def outline_thickness(self, value):
        self.uniform_buffer.data["outline_thickness"] = max(0.0, min(0.5, float(value)))
        self.uniform_buffer.update_full()

    @property
    def outline_color(self):
        """The color of the outline of the text."""
        return Color(self.uniform_buffer.data["outline_color"])

    @outline_color.setter
    def outline_color(self, outline_color):
        outline_color = Color(outline_color)
        self.uniform_buffer.data["outline_color"] = outline_color
        self.uniform_buffer.update_full()
        self._store.outline_color_is_transparent = outline_color.a < 1

    @property
    def outline_color_is_transparent(self):
        """Whether the outline_color is (semi) transparent (i.e. not fully opaque)."""
        return self._store.outline_color_is_transparent

    @property
    def weight_offset(self):
        """A value representing an offset to the font weight. Font weights
        are in the range 100-900, so this value should be in the same order of
        magnitude. Can be negative to make text thinner. Default zero.

        One use-case is to make dark text on a light background 50 units
        thicker, to counter the psychological effect that such text
        *looks* thinner than a light text on a dark background.
        """
        return float(self.uniform_buffer.data["weight_offset"])

    @weight_offset.setter
    def weight_offset(self, value):
        self.uniform_buffer.data["weight_offset"] = float(value)
        self.uniform_buffer.update_full()
